# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""abstract class for http request"""

import time
import random
import base64
from hashlib import sha1  # pylint: disable=E0611
from calendar import timegm
from datetime import date, datetime
from warnings import warn
from io import BytesIO

from six import PY2, text_type, string_types
from six.moves import http_client
from six.moves.urllib.parse import urlsplit, quote as urlquote
from six.moves.http_cookies import SimpleCookie

from rql.utils import rqlvar_maker

from logilab.common.decorators import cached
from logilab.common.deprecation import deprecated

from cubicweb import AuthenticationError
from cubicweb.req import RequestSessionBase
from cubicweb.uilib import remove_html_tags, js
from cubicweb.utils import HTMLHead, make_uid
from cubicweb.web import (INTERNAL_FIELD_VALUE, LOGGER, NothingToEdit,
                          RequestError, StatusResponse)
from cubicweb.web.httpcache import get_validators
from cubicweb.web.http_headers import Headers, Cookie


_MARKER = object()


def build_cb_uid(seed):
    sha = sha1(('%s%s%s' % (time.time(), seed, random.random())).encode('ascii'))
    return 'cb_%s' % (sha.hexdigest())


def list_form_param(form, param, pop=False):
    """get param from form parameters and return its value as a list,
    skipping internal markers if any

    * if the parameter isn't defined, return an empty list
    * if the parameter is a single (unicode) value, return a list
      containing that value
    * if the parameter is already a list or tuple, just skip internal
      markers

    if pop is True, the parameter is removed from the form dictionary
    """
    if pop:
        try:
            value = form.pop(param)
        except KeyError:
            return []
    else:
        value = form.get(param, ())
    if value is None:
        value = ()
    elif not isinstance(value, (list, tuple)):
        value = [value]
    return [v for v in value if v != INTERNAL_FIELD_VALUE]


class Counter(object):
    """A picklable counter object, usable for e.g. page tab index count"""
    __slots__ = ('value',)

    def __init__(self, initialvalue=0):
        self.value = initialvalue

    def __call__(self):
        value = self.value
        self.value += 1
        return value

    def __getstate__(self):
        return {'value': self.value}

    def __setstate__(self, state):
        self.value = state['value']


class _CubicWebRequestBase(RequestSessionBase):
    """abstract HTTP request, should be extended according to the HTTP backend
    Immutable attributes that describe the received query and generic configuration
    """
    ajax_request = False # to be set to True by ajax controllers

    def __init__(self, vreg, form=None, headers=None):
        """
        :vreg: Vregistry,
        :form: Forms value
        :headers: dict, request header
        """
        super(_CubicWebRequestBase, self).__init__(vreg)
        #: User interface property (see :ref:`uiprops`)
        self.uiprops = None
        #: url for serving datadir (see :ref:`resources`)
        self.datadir_url = None
        # some config (i.e. "pyramid") do not have "uiprops" nor "datadir_url"
        # attributes)
        self.uiprops = getattr(vreg.config, 'uiprops', None)
        self.datadir_url = getattr(vreg.config, 'datadir_url', None)
        #: enable UStringIO's write tracing
        self.tracehtml = False
        if vreg.config.debugmode:
            self.tracehtml = bool(form.pop('_cwtracehtml', False))
        #: raw html headers that can be added from any view
        self.html_headers = HTMLHead(self, tracewrites=self.tracehtml)
        #: received headers
        self._headers_in = Headers()
        if headers is not None:
            for k, v in headers.items():
                self._headers_in.addRawHeader(k, v)
        #: form parameters
        self.setup_params(form)
        #: received body
        self.content = BytesIO()
        # prepare output header
        #: Header used for the final response
        self.headers_out = Headers()
        #: HTTP status use by the final response
        self.status_out = 200
        # set up language based on site default (we don't have a user yet, and might not get one)
        self.set_language(vreg.property_value('ui.language'))
        #: dictionary that may be used to store request data that has to be
        #: shared among various components used to publish the request (views,
        #: controller, application...)
        self.data = {}
        self._search_state = None
        #: page id, set by htmlheader template
        self.pageid = None
        self._set_pageid()

    def _set_pageid(self):
        """initialize self.pageid
        if req.form provides a specific pageid, use it, otherwise build a
        new one.
        """
        pid = self.form.get('pageid')
        if pid is None:
            pid = make_uid(id(self))
            self.html_headers.define_var('pageid', pid, override=False)
        self.pageid = pid

    def _get_json_request(self):
        warn('[3.15] self._cw.json_request is deprecated, use self._cw.ajax_request instead',
             DeprecationWarning, stacklevel=2)
        return self.ajax_request
    def _set_json_request(self, value):
        warn('[3.15] self._cw.json_request is deprecated, use self._cw.ajax_request instead',
             DeprecationWarning, stacklevel=2)
        self.ajax_request = value
    json_request = property(_get_json_request, _set_json_request)

    @property
    def authmode(self):
        """Authentification mode of the instance
        (see :ref:`WebServerConfig`)"""
        return self.vreg.config['auth-mode']

    # Various variable generator.

    @property
    def varmaker(self):
        """the rql varmaker is exposed both as a property and as the
        set_varmaker function since we've two use cases:

        * accessing the req.varmaker property to get a new variable name

        * calling req.set_varmaker() to ensure a varmaker is set for later ajax
          calls sharing our .pageid
        """
        return self.set_varmaker()

    def set_varmaker(self):
        varmaker = self.get_page_data('rql_varmaker')
        if varmaker is None:
            varmaker = rqlvar_maker()
        self.set_page_data('rql_varmaker', varmaker)
        return varmaker

    # input form parameters management ########################################

    # common form parameters which should be protected against html values
    # XXX can't add 'eid' for instance since it may be multivalued
    # dont put rql as well, if query contains < and > it will be corrupted!
    no_script_form_params = set(('vid',
                                 'etype',
                                 'vtitle', 'title',
                                 '__redirectvid', '__redirectrql'))

    def setup_params(self, params):
        """WARNING: we're intentionally leaving INTERNAL_FIELD_VALUE here

        subclasses should overrides to
        """
        self.form = {}
        if params is None:
            return
        encoding = self.encoding
        for param, val in params.items():
            if isinstance(val, (tuple, list)):
                if PY2:
                    val = [unicode(x, encoding) for x in val]
                if len(val) == 1:
                    val = val[0]
            elif PY2 and isinstance(val, str):
                val = unicode(val, encoding)
            if param in self.no_script_form_params and val:
                val = self.no_script_form_param(param, val)
            if param == '_cwmsgid':
                self.set_message_id(val)
            else:
                self.form[param] = val

    def no_script_form_param(self, param, value):
        """ensure there is no script in a user form param

        by default return a cleaned string instead of raising a security
        exception

        this method should be called on every user input (form at least) fields
        that are at some point inserted in a generated html page to protect
        against script kiddies
        """
        # safety belt for strange urls like http://...?vtitle=yo&vtitle=yo
        if isinstance(value, (list, tuple)):
            self.error('no_script_form_param got a list (%s). Who generated the URL ?',
                       repr(value))
            value = value[0]
        return remove_html_tags(value)

    def list_form_param(self, param, form=None, pop=False):
        """get param from form parameters and return its value as a list,
        skipping internal markers if any

        * if the parameter isn't defined, return an empty list
        * if the parameter is a single (unicode) value, return a list
          containing that value
        * if the parameter is already a list or tuple, just skip internal
          markers

        if pop is True, the parameter is removed from the form dictionary
        """
        if form is None:
            form = self.form
        return list_form_param(form, param, pop)

    def reset_headers(self):
        """used by AutomaticWebTest to clear html headers between tests on
        the same resultset
        """
        self.html_headers = HTMLHead(self)
        return self

    # web state helpers #######################################################

    @property
    def message(self):
        try:
            return self.session.data.pop(self._msgid, u'')
        except AttributeError:
            try:
                return self._msg
            except AttributeError:
                return None

    def set_message(self, msg):
        assert isinstance(msg, text_type)
        self.reset_message()
        self._msg = msg

    def set_message_id(self, msgid):
        self._msgid = msgid

    @cached
    def redirect_message_id(self):
        return make_uid()

    def set_redirect_message(self, msg):
        # TODO - this should probably be merged with append_to_redirect_message
        assert isinstance(msg, text_type)
        msgid = self.redirect_message_id()
        self.session.data[msgid] = msg
        return msgid

    def append_to_redirect_message(self, msg):
        msgid = self.redirect_message_id()
        currentmsg = self.session.data.get(msgid)
        if currentmsg is not None:
            currentmsg = u'%s %s' % (currentmsg, msg)
        else:
            currentmsg = msg
        self.session.data[msgid] = currentmsg
        return msgid

    def reset_message(self):
        if hasattr(self, '_msg'):
            del self._msg
        if hasattr(self, '_msgid'):
            self.session.data.pop(self._msgid, u'')
            del self._msgid

    def _load_search_state(self, searchstate):
        if searchstate is None or searchstate == 'normal':
            self._search_state = ('normal',)
        else:
            self._search_state = ('linksearch', searchstate.split(':'))
            assert len(self._search_state[-1]) == 4, 'invalid searchstate'

    @property
    def search_state(self):
        """search state: 'normal' or 'linksearch' (i.e. searching for an object
        to create a relation with another)"""
        if self._search_state is None:
            searchstate = self.session.data.get('search_state', 'normal')
            self._load_search_state(searchstate)
        return self._search_state

    @search_state.setter
    def search_state(self, searchstate):
        self._search_state = searchstate

    def update_search_state(self):
        """update the current search state if needed"""
        searchstate = self.form.get('__mode')
        if searchstate:
            self.set_search_state(searchstate)

    def set_search_state(self, searchstate):
        """set a new search state"""
        self.session.data['search_state'] = searchstate
        self._load_search_state(searchstate)

    def match_search_state(self, rset):
        """when searching an entity to create a relation, return True if entities in
        the given rset may be used as relation end
        """
        try:
            searchedtype = self.search_state[1][-1]
        except IndexError:
            return False # no searching for association
        for etype in rset.column_types(0):
            if etype != searchedtype:
                return False
        return True

    # web edition helpers #####################################################

    @cached # so it's writed only once
    def fckeditor_config(self):
        fckeditor_url = self.build_url('fckeditor/fckeditor.js')
        self.add_js(fckeditor_url, localfile=False)
        self.html_headers.define_var('fcklang', self.lang)
        self.html_headers.define_var('fckconfigpath',
                                     self.data_url('cubicweb.fckcwconfig.js'))
    def use_fckeditor(self):
        return self.vreg.config.fckeditor_installed() and self.property_value('ui.fckeditor')

    def edited_eids(self, withtype=False):
        """return a list of edited eids"""
        yielded = False
        # warning: use .keys since the caller may change `form`
        form = self.form
        try:
            eids = form['eid']
        except KeyError:
            raise NothingToEdit(self._('no selected entities'))
        if isinstance(eids, string_types):
            eids = (eids,)
        for peid in eids:
            if withtype:
                typekey = '__type:%s' % peid
                assert typekey in form, 'no entity type specified'
                yield peid, form[typekey]
            else:
                yield peid
            yielded = True
        if not yielded:
            raise NothingToEdit(self._('no selected entities'))

    # minparams=3 by default: at least eid, __type, and some params to change
    def extract_entity_params(self, eid, minparams=3):
        """extract form parameters relative to the given eid"""
        params = {}
        eid = str(eid)
        form = self.form
        for param in form:
            try:
                name, peid = param.split(':', 1)
            except ValueError:
                if not param.startswith('__') and param not in ('eid', '_cw_fields'):
                    self.warning('param %s mis-formatted', param)
                continue
            if peid == eid:
                value = form[param]
                if value == INTERNAL_FIELD_VALUE:
                    value = None
                params[name] = value
        params['eid'] = eid
        if len(params) < minparams:
            raise RequestError(self._('missing parameters for entity %s') % eid)
        return params

    # XXX this should go to the GenericRelationsField. missing edition cancel protocol.

    def remove_pending_operations(self):
        """shortcut to clear req's pending_{delete,insert} entries

        This is needed when the edition is completed (whether it's validated
        or cancelled)
        """
        self.session.data.pop('pending_insert', None)
        self.session.data.pop('pending_delete', None)

    def cancel_edition(self, errorurl):
        """remove pending operations and `errorurl`'s specific stored data
        """
        self.session.data.pop(errorurl, None)
        self.remove_pending_operations()

    # high level methods for HTTP headers management ##########################

    # must be cached since login/password are popped from the form dictionary
    # and this method may be called multiple times during authentication
    @cached
    def get_authorization(self):
        """Parse and return the Authorization header"""
        if self.authmode == "cookie":
            try:
                user = self.form.pop("__login")
                passwd = self.form.pop("__password", '')
                return user, passwd.encode('UTF8')
            except KeyError:
                self.debug('no login/password in form params')
                return None, None
        else:
            return self.header_authorization()

    def get_cookie(self):
        """retrieve request cookies, returns an empty cookie if not found"""
        # XXX use http_headers implementation
        try:
            return SimpleCookie(self.get_header('Cookie'))
        except KeyError:
            return SimpleCookie()

    def set_cookie(self, name, value, maxage=300, expires=None, secure=False, httponly=False):
        """set / update a cookie

        by default, cookie will be available for the next 5 minutes.
        Give maxage = None to have a "session" cookie expiring when the
        client close its browser
        """
        if isinstance(name, SimpleCookie):
            warn('[3.13] set_cookie now takes name and value as two first '
                 'argument, not anymore cookie object and name',
                 DeprecationWarning, stacklevel=2)
            secure = name[value]['secure']
            name, value = value, name[value].value
        if maxage: # don't check is None, 0 may be specified
            assert expires is None, 'both max age and expires cant be specified'
            expires = maxage + time.time()
        elif expires:
            # we don't want to handle times before the EPOCH (cause bug on
            # windows). Also use > and not >= else expires == 0 and Cookie think
            # that means no expire...
            assert expires > date(1970, 1, 1)
            expires = timegm(expires.timetuple())
        else:
            expires = None
        # make sure cookie is set on the correct path
        cookie = Cookie(str(name), str(value), self.base_url_path(),
                        expires=expires, secure=secure, httponly=httponly)
        self.headers_out.addHeader('Set-cookie', cookie)

    def remove_cookie(self, name, bwcompat=None):
        """remove a cookie by expiring it"""
        if bwcompat is not None:
            warn('[3.13] remove_cookie now take only a name as argument',
                 DeprecationWarning, stacklevel=2)
            name = bwcompat
        self.set_cookie(name, '', maxage=0, expires=date(2000, 1, 1))

    def set_content_type(self, content_type, filename=None, encoding=None,
                         disposition='inline'):
        """set output content type for this request. An optional filename
        may be given.

        The disposition argument may be `attachement` or `inline` as specified
        for the Content-disposition HTTP header. The disposition parameter have
        no effect if no filename are specified.
        """
        if content_type.startswith('text/') and ';charset=' not in content_type:
            content_type += ';charset=' + (encoding or self.encoding)
        self.set_header('content-type', content_type)
        if filename:
            header = [disposition]
            unicode_filename = None
            try:
                ascii_filename = filename.encode('ascii').decode('ascii')
            except UnicodeEncodeError:
                # fallback filename for very old browser
                unicode_filename = filename
                ascii_filename = filename.encode('ascii', 'ignore').decode('ascii')
            # escape " and \
            # see http://greenbytes.de/tech/tc2231/#attwithfilenameandextparamescaped
            ascii_filename = ascii_filename.replace('\x5c', r'\\').replace('"', r'\"')
            header.append('filename="%s"' % ascii_filename)
            if unicode_filename is not None:
                # encoded filename according RFC5987
                urlquoted_filename = urlquote(unicode_filename.encode('utf-8'), '')
                header.append("filename*=utf-8''" + urlquoted_filename)
            self.set_header('content-disposition', ';'.join(header))

    # high level methods for HTML headers management ##########################

    def add_onload(self, jscode):
        self.html_headers.add_onload(jscode)

    def add_js(self, jsfiles, localfile=True):
        """specify a list of JS files to include in the HTML headers.

        :param jsfiles: a JS filename or a list of JS filenames
        :param localfile: if True, the default data dir prefix is added to the
                          JS filename
        """
        if isinstance(jsfiles, string_types):
            jsfiles = (jsfiles,)
        for jsfile in jsfiles:
            if localfile:
                jsfile = self.data_url(jsfile)
            self.html_headers.add_js(jsfile)

    def add_css(self, cssfiles, media=u'all', localfile=True, ieonly=False,
                iespec=u'[if lt IE 8]'):
        """specify a CSS file to include in the HTML headers

        :param cssfiles: a CSS filename or a list of CSS filenames.
        :param media: the CSS's media if necessary
        :param localfile: if True, the default data dir prefix is added to the
                          CSS filename
        :param ieonly: True if this css is specific to IE
        :param iespec: conditional expression that will be used around
                       the css inclusion. cf:
                       http://msdn.microsoft.com/en-us/library/ms537512(VS.85).aspx
        """
        if isinstance(cssfiles, string_types):
            cssfiles = (cssfiles,)
        if ieonly:
            if self.ie_browser():
                extraargs = [iespec]
                add_css = self.html_headers.add_ie_css
            else:
                return # no need to do anything on non IE browsers
        else:
            extraargs = []
            add_css = self.html_headers.add_css
        for cssfile in cssfiles:
            if localfile:
                cssfile = self.data_url(cssfile)
            add_css(cssfile, media, *extraargs)

    def ajax_replace_url(self, nodeid, replacemode='replace', **extraparams):
        """builds an ajax url that will replace nodeid's content

        :param nodeid: the dom id of the node to replace
        :param replacemode: defines how the replacement should be done.

          Possible values are :
          - 'replace' to replace the node's content with the generated HTML
          - 'swap' to replace the node itself with the generated HTML
          - 'append' to append the generated HTML to the node's content

        Arbitrary extra named arguments may be given, they will be included as
        parameters of the generated url.
        """
        # define a function in headers and use it in the link to avoid url
        # unescaping pb: browsers give the js expression to the interpreter
        # after having url unescaping the content. This may make appear some
        # quote or other special characters that will break the js expression.
        extraparams.setdefault('fname', 'view')
        # remove pageid from the generated URL as it's forced as a parameter
        # to the loadxhtml call below.
        extraparams.pop('pageid', None)
        url = self.build_url('ajax', **extraparams)
        cbname = build_cb_uid(url[:50])
        # think to propagate pageid. XXX see https://www.cubicweb.org/ticket/1753121
        jscode = u'function %s() { $("#%s").%s; }' % (
            cbname, nodeid, js.loadxhtml(url, {'pageid': self.pageid},
                                         'get', replacemode))
        self.html_headers.add_post_inline_script(jscode)
        return "javascript: %s()" % cbname

    # urls/path management ####################################################

    def build_url(self, *args, **kwargs):
        """return an absolute URL using params dictionary key/values as URL
        parameters. Values are automatically URL quoted, and the
        publishing method to use may be specified or will be guessed.
        """
        if '__message' in kwargs:
            msg = kwargs.pop('__message')
            kwargs['_cwmsgid'] = self.set_redirect_message(msg)
        if not args:
            method = 'view'
            if (self.from_controller() == 'view'
                and not '_restpath' in kwargs):
                method = self.relative_path(includeparams=False) or 'view'
            args = (method,)
        return super(_CubicWebRequestBase, self).build_url(*args, **kwargs)

    def build_url_path(self, *args):
        path = super(_CubicWebRequestBase, self).build_url_path(*args)
        lang_prefix = ''
        if self.lang is not None and self.vreg.config.get('language-mode') == 'url-prefix':
            lang_prefix = '%s/' % self.lang
        return lang_prefix + path

    def url(self, includeparams=True):
        """return currently accessed url"""
        return self.base_url() + self.relative_path(includeparams)

    def selected(self, url):
        """return True if the url is equivalent to currently accessed url"""
        reqpath = self.relative_path().lower()
        baselen = len(self.base_url())
        return (reqpath == url[baselen:].lower())

    def base_url_prepend_host(self, hostname):
        protocol, roothost = urlsplit(self.base_url())[:2]
        if roothost.startswith('www.'):
            roothost = roothost[4:]
        return '%s://%s.%s' % (protocol, hostname, roothost)

    def base_url_path(self):
        """returns the absolute path of the base url"""
        return urlsplit(self.base_url())[2]

    def data_url(self, relpath):
        """returns the absolute path for a data resource"""
        return self.datadir_url + relpath

    @cached
    def from_controller(self):
        """return the id (string) of the controller issuing the request"""
        controller = self.relative_path(False).split('/', 1)[0]
        if controller in self.vreg['controllers']:
            return controller
        return 'view'

    def is_client_cache_valid(self):
        """check if a client cached page exists (as specified in request
        headers) and is still usable.

        Return False if the page has to be calculated, else True.

        Some response cache headers may be set by this method.
        """
        modified = True
        # Here, we search for any invalid 'not modified' condition
        # see http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html#sec13.3
        validators = get_validators(self._headers_in)
        if validators: # if we have no
            modified = any(func(val, self.headers_out) for func, val in validators)
        # Forge expected response
        if not modified:
            # overwrite headers_out to forge a brand new not-modified response
            self.headers_out = self._forge_cached_headers()
            if self.http_method() in ('HEAD', 'GET'):
                self.status_out = http_client.NOT_MODIFIED
            else:
                self.status_out = http_client.PRECONDITION_FAILED
            # XXX replace by True once validate_cache bw compat method is dropped
            return self.status_out
        # XXX replace by False once validate_cache bw compat method is dropped
        return None

    @deprecated('[3.18] use .is_client_cache_valid() method instead')
    def validate_cache(self):
        """raise a `StatusResponse` exception if a cached page along the way
        exists and is still usable.
        """
        status_code = self.is_client_cache_valid()
        if status_code is not None:
            raise StatusResponse(status_code)

    # abstract methods to override according to the web front-end #############

    def http_method(self):
        """returns 'POST', 'GET', 'HEAD', etc."""
        raise NotImplementedError()

    def _forge_cached_headers(self):
        # overwrite headers_out to forge a brand new not-modified response
        headers = Headers()
        for header in (
            # Required from sec 10.3.5:
            'date', 'etag', 'content-location', 'expires',
            'cache-control', 'vary',
            # Others:
            'server', 'proxy-authenticate', 'www-authenticate', 'warning'):
            value = self.headers_out.getRawHeaders(header)
            if value is not None:
                headers.setRawHeaders(header, value)
        return headers

    def relative_path(self, includeparams=True):
        """return the normalized path of the request (ie at least relative
        to the instance's root, but some other normalization may be needed
        so that the returned path may be used to compare to generated urls

        :param includeparams:
           boolean indicating if GET form parameters should be kept in the path
        """
        raise NotImplementedError()

    # http headers ############################################################

    ### incoming headers

    def get_header(self, header, default=None, raw=True):
        """return the value associated with the given input header, raise
        KeyError if the header is not set
        """
        if raw:
            return self._headers_in.getRawHeaders(header, [default])[0]
        return self._headers_in.getHeader(header, default)

    def header_accept_language(self):
        """returns an ordered list of preferred languages"""
        acceptedlangs = self.get_header('Accept-Language', raw=False) or {}
        for lang, _ in sorted(acceptedlangs.items(), key=lambda x: x[1],
                              reverse=True):
            lang = lang.split('-')[0]
            yield lang

    def header_if_modified_since(self):
        """If the HTTP header If-modified-since is set, return the equivalent
        date time value (GMT), else return None
        """
        mtime = self.get_header('If-modified-since', raw=False)
        if mtime:
            return datetime.utcfromtimestamp(mtime)
        return None

    ### outcoming headers
    def set_header(self, header, value, raw=True):
        """set an output HTTP header"""
        if raw:
            # adding encoded header is important, else page content
            # will be reconverted back to unicode and apart unefficiency, this
            # may cause decoding problem (e.g. when downloading a file)
            self.headers_out.setRawHeaders(header, [str(value)])
        else:
            self.headers_out.setHeader(header, value)

    def add_header(self, header, value):
        """add an output HTTP header"""
        # adding encoded header is important, else page content
        # will be reconverted back to unicode and apart unefficiency, this
        # may cause decoding problem (e.g. when downloading a file)
        self.headers_out.addRawHeader(header, str(value))

    def remove_header(self, header):
        """remove an output HTTP header"""
        self.headers_out.removeHeader(header)

    def header_authorization(self):
        """returns a couple (auth-type, auth-value)"""
        auth = self.get_header("Authorization", None)
        if auth:
            scheme, rest = auth.split(' ', 1)
            scheme = scheme.lower()
            try:
                assert scheme == "basic"
                user, passwd = base64.decodestring(rest.encode('ascii')).split(b":", 1)
                # XXX HTTP header encoding: use email.Header?
                return user.decode('UTF8'), passwd
            except Exception as ex:
                self.debug('bad authorization %s (%s: %s)',
                           auth, ex.__class__.__name__, ex)
        return None, None

    def parse_accept_header(self, header):
        """returns an ordered list of accepted values"""
        try:
            value_parser, value_sort_key = ACCEPT_HEADER_PARSER[header.lower()]
        except KeyError:
            value_parser = value_sort_key = None
        accepteds = self.get_header(header, '')
        values = _parse_accept_header(accepteds, value_parser, value_sort_key)
        return (raw_value for (raw_value, parsed_value, score) in values)

    @deprecated('[3.17] demote_to_html is deprecated as we always serve html')
    def demote_to_html(self):
        """helper method to dynamically set request content type to text/html

        The global doctype and xmldec must also be changed otherwise the browser
        will display '<[' at the beginning of the page
        """
        pass


    # xml doctype #############################################################

    def set_doctype(self, doctype, reset_xmldecl=None):
        """helper method to dynamically change page doctype

        :param doctype: the new doctype, e.g. '<!DOCTYPE html>'
        """
        if reset_xmldecl is not None:
            warn('[3.17] reset_xmldecl is deprecated as we only serve html',
                 DeprecationWarning, stacklevel=2)
        self.main_stream.set_doctype(doctype)

    # page data management ####################################################

    def get_page_data(self, key, default=None):
        """return value associated to `key` in current page data"""
        page_data = self.session.data.get(self.pageid)
        if page_data is None:
            return default
        return page_data.get(key, default)

    def set_page_data(self, key, value):
        """set value associated to `key` in current page data"""
        self.html_headers.add_unload_pagedata()
        page_data = self.session.data.setdefault(self.pageid, {})
        page_data[key] = value
        self.session.data[self.pageid] = page_data

    def del_page_data(self, key=None):
        """remove value associated to `key` in current page data
        if `key` is None, all page data will be cleared
        """
        if key is None:
            self.session.data.pop(self.pageid, None)
        else:
            try:
                page_data = self.session.data[self.pageid]
                del page_data[key]
            except KeyError:
                pass
            else:
                # make sure we write the session data value in the
                # self.session.data dict-like object so any session
                # handler can "detect" and manage the persistency
                self.session.data[self.pageid] = page_data

    # user-agent detection ####################################################

    @cached
    def useragent(self):
        return self.get_header('User-Agent', None)

    def ie_browser(self):
        useragent = self.useragent()
        return useragent and 'MSIE' in useragent

    @deprecated('[3.17] xhtml_browser is deprecated (xhtml is no longer served)')
    def xhtml_browser(self):
        """return True if the browser is considered as xhtml compatible.

        If the instance is configured to always return text/html and not
        application/xhtml+xml, this method will always return False, even though
        this is semantically different
        """
        return False

    def html_content_type(self):
        return 'text/html'

    def negotiated_language(self):
        self.headers_out.addHeader('Vary', 'Accept-Language')
        for lang in self.header_accept_language():
            if lang in self.translations:
                return lang
        return None


def _cnx_func(name):
    def proxy(req, *args, **kwargs):
        return getattr(req.cnx, name)(*args, **kwargs)
    return proxy

class _NeedAuthAccessMock(object):

    def __getattribute__(self, attr):
        raise AuthenticationError()

    def __bool__(self):
        return False

    __nonzero__ = __bool__

class _MockAnonymousSession(object):
    sessionid = 'thisisnotarealsession'

    @property
    def data(self):
        return {}

    @property
    def anonymous_session(self):
        return True

class ConnectionCubicWebRequestBase(_CubicWebRequestBase):
    cnx = None
    session = None

    def __init__(self, vreg, form=None, headers={}):
        """"""
        self.vreg = vreg
        try:
            # no vreg or config which doesn't handle translations
            self.translations = vreg.config.translations
        except AttributeError:
            self.translations = {}
        super(ConnectionCubicWebRequestBase, self).__init__(vreg, form=form, headers=headers)
        self.session = _MockAnonymousSession()
        self.cnx = self.user = _NeedAuthAccessMock()

    @property
    def transaction_data(self):
        return self.cnx.transaction_data

    def set_cnx(self, cnx):
        self.cnx = cnx
        self.session = cnx.session
        self._set_user(cnx.user)
        # set language according to the one defined on the connection which consider user's
        # preference
        self.set_language(cnx.lang)

    def execute(self, *args, **kwargs):
        rset = self.cnx.execute(*args, **kwargs)
        rset.req = self
        return rset

    entity_metas = _cnx_func('entity_metas')  # XXX deprecated
    entity_type = _cnx_func('entity_type')
    source_defs = _cnx_func('source_defs')
    get_shared_data = _cnx_func('get_shared_data')
    set_shared_data = _cnx_func('set_shared_data')

    # security #################################################################

    security_enabled = _cnx_func('security_enabled')

    # server-side service call #################################################

    def call_service(self, regid, **kwargs):
        return self.cnx.call_service(regid, **kwargs)

    # entities cache management ###############################################

    def entity_cache(self, eid):
        return self.transaction_data['req_ecache'][eid]

    def set_entity_cache(self, entity):
        ecache = self.transaction_data.setdefault('req_ecache', {})
        ecache.setdefault(entity.eid, entity)

    def cached_entities(self):
        return self.transaction_data.get('req_ecache', {}).values()

    def drop_entity_cache(self):
        self.transaction_data.pop('req_ecache', None)


CubicWebRequestBase = ConnectionCubicWebRequestBase


## HTTP-accept parsers / utilies ##############################################
def _mimetype_sort_key(accept_info):
    """accepted mimetypes must be sorted by :

    1/ highest score first
    2/ most specific mimetype first, e.g. :
       - 'text/html level=1' is more specific 'text/html'
       - 'text/html' is more specific than 'text/*'
       - 'text/*' itself more specific than '*/*'

    """
    raw_value, (media_type, media_subtype, media_type_params), score = accept_info
    # FIXME: handle '+' in media_subtype ? (should xhtml+xml have a
    # higher precedence than xml ?)
    if media_subtype == '*':
        score -= 0.0001
    if media_type == '*':
        score -= 0.0001
    return 1./score, media_type, media_subtype, 1./(1+len(media_type_params))

def _charset_sort_key(accept_info):
    """accepted mimetypes must be sorted by :

    1/ highest score first
    2/ most specific charset first, e.g. :
       - 'utf-8' is more specific than '*'
    """
    raw_value, value, score = accept_info
    if value == '*':
        score -= 0.0001
    return 1./score, value

def _parse_accept_header(raw_header, value_parser=None, value_sort_key=None):
    """returns an ordered list accepted types

    :param value_parser: a function to parse a raw accept chunk. If None
    is provided, the function defaults to identity. If a function is provided,
    it must accept 2 parameters ``value`` and ``other_params``. ``value`` is
    the value found before the first ';', `other_params` is a dictionary
    built from all other chunks after this first ';'

    :param value_sort_key: a key function to sort values found in the accept
    header. This function will be passed a 3-tuple
    (raw_value, parsed_value, score). If None is provided, the default
    sort_key is 1./score

    :return: a list of 3-tuple (raw_value, parsed_value, score),
    ordered by score. ``parsed_value`` will be the return value of
    ``value_parser(raw_value)``
    """
    if value_sort_key is None:
        value_sort_key = lambda infos: 1./infos[-1]
    values = []
    for info in raw_header.split(','):
        score = 1.0
        other_params = {}
        try:
            value, infodef = info.split(';', 1)
        except ValueError:
            value = info
        else:
            for info in infodef.split(';'):
                try:
                    infokey, infoval = info.split('=')
                    if infokey == 'q': # XXX 'level'
                        score = float(infoval)
                        continue
                except ValueError:
                    continue
                other_params[infokey] = infoval
        parsed_value = value_parser(value, other_params) if value_parser else value
        values.append( (value.strip(), parsed_value, score) )
    values.sort(key=value_sort_key)
    return values


def _mimetype_parser(value, other_params):
    """return a 3-tuple
    (type, subtype, type_params) corresponding to the mimetype definition
    e.g. : for 'text/*', `mimetypeinfo` will be ('text', '*', {}), for
    'text/html;level=1', `mimetypeinfo` will be ('text', '*', {'level': '1'})
    """
    try:
        media_type, media_subtype = value.strip().split('/', 1)
    except ValueError: # safety belt : '/' should always be present
        media_type = value.strip()
        media_subtype = '*'
    return (media_type, media_subtype, other_params)


ACCEPT_HEADER_PARSER = {
    'accept': (_mimetype_parser, _mimetype_sort_key),
    'accept-charset': (None, _charset_sort_key),
    }

from cubicweb import set_log_methods
set_log_methods(_CubicWebRequestBase, LOGGER)



# copyright 2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unittest for cubicweb.devtools.httptest module"""

from six.moves import http_client

from logilab.common.testlib import Tags
from cubicweb.devtools.httptest import CubicWebServerTC, CubicWebWsgiTC


class TwistedCWAnonTC(CubicWebServerTC):

    def test_response(self):
        try:
            response = self.web_get()
        except http_client.NotConnected as ex:
            self.fail("Can't connection to test server: %s" % ex)

    def test_response_anon(self):
        response = self.web_get()
        self.assertEqual(response.status, http_client.OK)

    def test_base_url(self):
        if self.config['base-url'] not in self.web_get().read().decode('ascii'):
            self.fail('no mention of base url in retrieved page')


class TwistedCWIdentTC(CubicWebServerTC):
    test_db_id = 'httptest-cwident'
    anonymous_allowed = False
    tags = CubicWebServerTC.tags | Tags(('auth',))

    def test_response_denied(self):
        response = self.web_get()
        self.assertEqual(response.status, http_client.FORBIDDEN)

    def test_login(self):
        response = self.web_get()
        if response.status != http_client.FORBIDDEN:
            self.skipTest('Already authenticated, "test_response_denied" must have failed')
        # login
        self.web_login(self.admlogin, self.admpassword)
        response = self.web_get()
        self.assertEqual(response.status, http_client.OK, response.body)
        # logout
        self.web_logout()
        response = self.web_get()
        self.assertEqual(response.status, http_client.FORBIDDEN, response.body)


class WsgiCWAnonTC(CubicWebWsgiTC):

    def test_response(self):
        try:
            response = self.web_get()
        except http_client.NotConnected as ex:
            self.fail("Can't connection to test server: %s" % ex)

    def test_response_anon(self):
        response = self.web_get()
        self.assertEqual(response.status, http_client.OK)

    def test_base_url(self):
        if self.config['base-url'] not in self.web_get().read().decode('ascii'):
            self.fail('no mention of base url in retrieved page')


class WsgiCWIdentTC(CubicWebWsgiTC):
    test_db_id = 'httptest-cwident'
    anonymous_allowed = False
    tags = CubicWebServerTC.tags | Tags(('auth',))

    def test_response_denied(self):
        response = self.web_get()
        self.assertEqual(response.status, http_client.FORBIDDEN)

    def test_login(self):
        response = self.web_get()
        if response.status != http_client.FORBIDDEN:
            self.skipTest('Already authenticated, "test_response_denied" must have failed')
        # login
        self.web_login(self.admlogin, self.admpassword)
        response = self.web_get()
        self.assertEqual(response.status, http_client.OK, response.body)
        # logout
        self.web_logout()
        response = self.web_get()
        self.assertEqual(response.status, http_client.FORBIDDEN, response.body)


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from cubicweb import ValidationError
from cubicweb.devtools.testlib import CubicWebTC

def add_wf(shell, etype, name=None, default=False):
    if name is None:
        name = etype
    return shell.add_workflow(name, etype, default=default,
                              ensure_workflowable=False)

def parse_hist(wfhist):
    return [(ti.previous_state.name, ti.new_state.name,
             ti.transition and ti.transition.name, ti.comment)
            for ti in wfhist]


class WorkflowBuildingTC(CubicWebTC):

    def test_wf_construction(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'Company')
            foo = wf.add_state(u'foo', initial=True)
            bar = wf.add_state(u'bar')
            self.assertEqual(wf.state_by_name('bar').eid, bar.eid)
            self.assertEqual(wf.state_by_name('barrr'), None)
            baz = wf.add_transition(u'baz', (foo,), bar, ('managers',))
            self.assertEqual(wf.transition_by_name('baz').eid, baz.eid)
            self.assertEqual(len(baz.require_group), 1)
            self.assertEqual(baz.require_group[0].name, 'managers')

    def test_duplicated_state(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'Company')
            wf.add_state(u'foo', initial=True)
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                wf.add_state(u'foo')
            self.assertEqual({'name': u'%(KEY-rtype)s is part of violated unicity constraint',
                              'state_of': u'%(KEY-rtype)s is part of violated unicity constraint',
                              '': u'some relations violate a unicity constraint'},
                             cm.exception.errors)
            shell.rollback()
            # no pb if not in the same workflow
            wf2 = add_wf(shell, 'Company')
            foo = wf2.add_state(u'foo', initial=True)
            shell.commit()
            # gnark gnark
            bar = wf.add_state(u'bar')
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                bar.cw_set(name=u'foo')
            shell.rollback()
            self.assertEqual({'name': u'%(KEY-rtype)s is part of violated unicity constraint',
                              'state_of': u'%(KEY-rtype)s is part of violated unicity constraint',
                              '': u'some relations violate a unicity constraint'},
                             cm.exception.errors)

    def test_duplicated_transition(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'Company')
            foo = wf.add_state(u'foo', initial=True)
            bar = wf.add_state(u'bar')
            wf.add_transition(u'baz', (foo,), bar, ('managers',))
            with self.assertRaises(ValidationError) as cm:
                wf.add_transition(u'baz', (bar,), foo)
            self.assertEqual({'name': u'%(KEY-rtype)s is part of violated unicity constraint',
                              'transition_of': u'%(KEY-rtype)s is part of violated unicity constraint',
                              '': u'some relations violate a unicity constraint'},
                             cm.exception.errors)
            shell.rollback()
            # no pb if not in the same workflow
            wf2 = add_wf(shell, 'Company')
            foo = wf2.add_state(u'foo', initial=True)
            bar = wf2.add_state(u'bar')
            wf2.add_transition(u'baz', (foo,), bar, ('managers',))
            shell.commit()
            # gnark gnark
            biz = wf2.add_transition(u'biz', (bar,), foo)
            shell.commit()
            with self.assertRaises(ValidationError) as cm:
                biz.cw_set(name=u'baz')
            shell.rollback()
            self.assertEqual({'name': u'%(KEY-rtype)s is part of violated unicity constraint',
                              'transition_of': u'%(KEY-rtype)s is part of violated unicity constraint',
                              '': u'some relations violate a unicity constraint'},
                             cm.exception.errors)


class WorkflowTC(CubicWebTC):

    def setup_database(self):
        rschema = self.schema['in_state']
        for rdef in rschema.rdefs.values():
            self.assertEqual(rdef.cardinality, '1*')
        with self.admin_access.client_cnx() as cnx:
            self.member_eid = self.create_user(cnx, 'member').eid
            cnx.commit()

    def test_workflow_base(self):
        with self.admin_access.web_request() as req:
            e = self.create_user(req, 'toto')
            iworkflowable = e.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.state, 'activated')
            iworkflowable.change_state('deactivated', u'deactivate 1')
            req.cnx.commit()
            iworkflowable.change_state('activated', u'activate 1')
            req.cnx.commit()
            iworkflowable.change_state('deactivated', u'deactivate 2')
            req.cnx.commit()
            e.cw_clear_relation_cache('wf_info_for', 'object')
            self.assertEqual([tr.comment for tr in e.reverse_wf_info_for],
                              ['deactivate 1', 'activate 1', 'deactivate 2'])
            self.assertEqual(iworkflowable.latest_trinfo().comment, 'deactivate 2')

    def test_possible_transitions(self):
        with self.admin_access.web_request() as req:
            user = req.execute('CWUser X').get_entity(0, 0)
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            trs = list(iworkflowable.possible_transitions())
            self.assertEqual(len(trs), 1)
            self.assertEqual(trs[0].name, u'deactivate')
            self.assertEqual(trs[0].destination(None).name, u'deactivated')
        # test a std user get no possible transition
        with self.new_access('member').web_request() as req:
            # fetch the entity using the new session
            trs = list(req.user.cw_adapt_to('IWorkflowable').possible_transitions())
            self.assertEqual(len(trs), 0)

    def _test_manager_deactivate(self, user):
        iworkflowable = user.cw_adapt_to('IWorkflowable')
        user.cw_clear_relation_cache('in_state', 'subject')
        self.assertEqual(len(user.in_state), 1)
        self.assertEqual(iworkflowable.state, 'deactivated')
        trinfo = iworkflowable.latest_trinfo()
        self.assertEqual(trinfo.previous_state.name, 'activated')
        self.assertEqual(trinfo.new_state.name, 'deactivated')
        self.assertEqual(trinfo.comment, 'deactivate user')
        self.assertEqual(trinfo.comment_format, 'text/plain')
        return trinfo

    def test_change_state(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            iworkflowable.change_state('deactivated', comment=u'deactivate user')
            trinfo = self._test_manager_deactivate(user)
            self.assertEqual(trinfo.transition, None)

    def test_set_in_state_bad_wf(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            s = wf.add_state(u'foo', initial=True)
            shell.commit()
        with self.admin_access.repo_cnx() as cnx:
            with cnx.security_enabled(write=False):
                with self.assertRaises(ValidationError) as cm:
                    cnx.execute('SET X in_state S WHERE X eid %(x)s, S eid %(s)s',
                                {'x': cnx.user.eid, 's': s.eid})
                self.assertEqual(cm.exception.errors, {'in_state-subject': "state doesn't belong to entity's workflow. "
                                          "You may want to set a custom workflow for this entity first."})

    def test_fire_transition(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('deactivate', comment=u'deactivate user')
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, 'deactivated')
            self._test_manager_deactivate(user)
            trinfo = self._test_manager_deactivate(user)
            self.assertEqual(trinfo.transition.name, 'deactivate')

    def test_goback_transition(self):
        with self.admin_access.web_request() as req:
            wf = req.user.cw_adapt_to('IWorkflowable').current_workflow
            asleep = wf.add_state('asleep')
            wf.add_transition('rest', (wf.state_by_name('activated'),
                                       wf.state_by_name('deactivated')),
                              asleep)
            wf.add_transition('wake up', asleep)
            user = self.create_user(req, 'stduser')
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('rest')
            req.cnx.commit()
            iworkflowable.fire_transition('wake up')
            req.cnx.commit()
            self.assertEqual(iworkflowable.state, 'activated')
            iworkflowable.fire_transition('deactivate')
            req.cnx.commit()
            iworkflowable.fire_transition('rest')
            req.cnx.commit()
            iworkflowable.fire_transition('wake up')
            req.cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, 'deactivated')

    # XXX test managers can change state without matching transition

    def _test_stduser_deactivate(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, 'tutu')
        with self.new_access('tutu').web_request() as req:
            iworkflowable = req.entity_from_eid(self.member_eid).cw_adapt_to('IWorkflowable')
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition('deactivate')
            self.assertEqual(cm.exception.errors, {'by_transition-subject': "transition may not be fired"})
        with self.new_access('member').web_request() as req:
            iworkflowable = req.entity_from_eid(self.member_eid).cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('deactivate')
            req.cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition('activate')
            self.assertEqual(cm.exception.errors, {'by_transition-subject': "transition may not be fired"})

    def test_fire_transition_owned_by(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('INSERT RQLExpression X: X exprtype "ERQLExpression", '
                        'X expression "X owned_by U", T condition X '
                        'WHERE T name "deactivate"')
            cnx.commit()
        self._test_stduser_deactivate()

    def test_fire_transition_has_update_perm(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('INSERT RQLExpression X: X exprtype "ERQLExpression", '
                        'X expression "U has_update_permission X", T condition X '
                        'WHERE T name "deactivate"')
            cnx.commit()
        self._test_stduser_deactivate()

    def test_swf_base(self):
        """subworkflow

        +-----------+  tr1   +-----------+
        | swfstate1 | ------>| swfstate2 |
        +-----------+        +-----------+
                  |  tr2  +-----------+
                  `------>| swfstate3 |
                          +-----------+

        main workflow

        +--------+  swftr1             +--------+
        | state1 | -------[swfstate2]->| state2 |
        +--------+     |               +--------+
                       |               +--------+
                       `-[swfstate3]-->| state3 |
                                       +--------+
        """
        # sub-workflow
        with self.admin_access.shell() as shell:
            swf = add_wf(shell, 'CWGroup', name='subworkflow')
            swfstate1 = swf.add_state(u'swfstate1', initial=True)
            swfstate2 = swf.add_state(u'swfstate2')
            swfstate3 = swf.add_state(u'swfstate3')
            tr1 = swf.add_transition(u'tr1', (swfstate1,), swfstate2)
            tr2 = swf.add_transition(u'tr2', (swfstate1,), swfstate3)
            # main workflow
            mwf = add_wf(shell, 'CWGroup', name='main workflow', default=True)
            state1 = mwf.add_state(u'state1', initial=True)
            state2 = mwf.add_state(u'state2')
            state3 = mwf.add_state(u'state3')
            swftr1 = mwf.add_wftransition(u'swftr1', swf, state1,
                                          [(swfstate2, state2), (swfstate3, state3)])
            swf.cw_clear_all_caches()
            self.assertEqual(swftr1.destination(None).eid, swfstate1.eid)
        # workflows built, begin test
        with self.admin_access.web_request() as req:
            group = req.create_entity('CWGroup', name=u'grp1')
            req.cnx.commit()
            iworkflowable = group.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.current_state.eid, state1.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.subworkflow_input_transition(), None)
            iworkflowable.fire_transition('swftr1', u'go')
            req.cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, swfstate1.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, swf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.subworkflow_input_transition().eid, swftr1.eid)
            iworkflowable.fire_transition('tr1', u'go')
            req.cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, state2.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.subworkflow_input_transition(), None)
            # force back to swfstate1 is impossible since we can't any more find
            # subworkflow input transition
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.change_state(swfstate1, u'gadget')
            self.assertEqual(cm.exception.errors, {'to_state-subject': "state doesn't belong to entity's workflow"})
            req.cnx.rollback()
            # force back to state1
            iworkflowable.change_state('state1', u'gadget')
            iworkflowable.fire_transition('swftr1', u'au')
            group.cw_clear_all_caches()
            iworkflowable.fire_transition('tr2', u'chapeau')
            req.cnx.commit()
            group.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_state.eid, state3.eid)
            self.assertEqual(iworkflowable.current_workflow.eid, mwf.eid)
            self.assertEqual(iworkflowable.main_workflow.eid, mwf.eid)
            self.assertListEqual(parse_hist(iworkflowable.workflow_history),
                                  [('state1', 'swfstate1', 'swftr1', 'go'),
                                   ('swfstate1', 'swfstate2', 'tr1', 'go'),
                                   ('swfstate2', 'state2', 'swftr1', 'exiting from subworkflow subworkflow'),
                                   ('state2', 'state1', None, 'gadget'),
                                   ('state1', 'swfstate1', 'swftr1', 'au'),
                                   ('swfstate1', 'swfstate3', 'tr2', 'chapeau'),
                                   ('swfstate3', 'state3', 'swftr1', 'exiting from subworkflow subworkflow'),
                                   ])

    def test_swf_exit_consistency(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            swf = add_wf(shell, 'CWGroup', name='subworkflow')
            swfstate1 = swf.add_state(u'swfstate1', initial=True)
            swfstate2 = swf.add_state(u'swfstate2')
            tr1 = swf.add_transition(u'tr1', (swfstate1,), swfstate2)
            # main workflow
            mwf = add_wf(shell, 'CWGroup', name='main workflow', default=True)
            state1 = mwf.add_state(u'state1', initial=True)
            state2 = mwf.add_state(u'state2')
            state3 = mwf.add_state(u'state3')
            mwf.add_wftransition(u'swftr1', swf, state1,
                                 [(swfstate2, state2), (swfstate2, state3)])
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(cm.exception.errors, {'subworkflow_exit-subject': u"can't have multiple exits on the same state"})

    def test_swf_fire_in_a_row(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            subwf = add_wf(shell, 'CWGroup', name='subworkflow')
            xsigning = subwf.add_state('xsigning', initial=True)
            xaborted = subwf.add_state('xaborted')
            xsigned = subwf.add_state('xsigned')
            xabort = subwf.add_transition('xabort', (xsigning,), xaborted)
            xsign = subwf.add_transition('xsign', (xsigning,), xsigning)
            xcomplete = subwf.add_transition('xcomplete', (xsigning,), xsigned,
                                             type=u'auto')
            # main workflow
            twf = add_wf(shell, 'CWGroup', name='mainwf', default=True)
            created    = twf.add_state(_('created'), initial=True)
            identified = twf.add_state(_('identified'))
            released   = twf.add_state(_('released'))
            closed   = twf.add_state(_('closed'))
            twf.add_wftransition(_('identify'), subwf, (created,),
                                 [(xsigned, identified), (xaborted, created)])
            twf.add_wftransition(_('release'), subwf, (identified,),
                                 [(xsigned, released), (xaborted, identified)])
            twf.add_wftransition(_('close'), subwf, (released,),
                                 [(xsigned, closed), (xaborted, released)])
            shell.commit()
        with self.admin_access.repo_cnx() as cnx:
            group = cnx.create_entity('CWGroup', name=u'grp1')
            cnx.commit()
            iworkflowable = group.cw_adapt_to('IWorkflowable')
            for trans in ('identify', 'release', 'close'):
                iworkflowable.fire_transition(trans)
                cnx.commit()


    def test_swf_magic_tr(self):
        with self.admin_access.shell() as shell:
            # sub-workflow
            subwf = add_wf(shell, 'CWGroup', name='subworkflow')
            xsigning = subwf.add_state('xsigning', initial=True)
            xaborted = subwf.add_state('xaborted')
            xsigned = subwf.add_state('xsigned')
            xabort = subwf.add_transition('xabort', (xsigning,), xaborted)
            xsign = subwf.add_transition('xsign', (xsigning,), xsigned)
            # main workflow
            twf = add_wf(shell, 'CWGroup', name='mainwf', default=True)
            created    = twf.add_state(_('created'), initial=True)
            identified = twf.add_state(_('identified'))
            released   = twf.add_state(_('released'))
            twf.add_wftransition(_('identify'), subwf, created,
                                 [(xaborted, None), (xsigned, identified)])
            twf.add_wftransition(_('release'), subwf, identified,
                                 [(xaborted, None)])
            shell.commit()
        with self.admin_access.web_request() as req:
            group = req.create_entity('CWGroup', name=u'grp1')
            req.cnx.commit()
            iworkflowable = group.cw_adapt_to('IWorkflowable')
            for trans, nextstate in (('identify', 'xsigning'),
                                     ('xabort', 'created'),
                                     ('identify', 'xsigning'),
                                     ('xsign', 'identified'),
                                     ('release', 'xsigning'),
                                     ('xabort', 'identified')
                                     ):
                iworkflowable.fire_transition(trans)
                req.cnx.commit()
                group.cw_clear_all_caches()
                self.assertEqual(iworkflowable.state, nextstate)

    def test_replace_state(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWGroup', name='groupwf', default=True)
            s_new = wf.add_state('new', initial=True)
            s_state1 = wf.add_state('state1')
            wf.add_transition('tr', (s_new,), s_state1)
            shell.commit()

        with self.admin_access.repo_cnx() as cnx:
            group = cnx.create_entity('CWGroup', name=u'grp1')
            cnx.commit()

            iwf = group.cw_adapt_to('IWorkflowable')
            iwf.fire_transition('tr')
            cnx.commit()
            group.cw_clear_all_caches()

            wf = cnx.entity_from_eid(wf.eid)
            wf.add_state('state2')
            with cnx.security_enabled(write=False):
                wf.replace_state('state1', 'state2')
            cnx.commit()

            self.assertEqual(iwf.state, 'state2')
            self.assertEqual(iwf.latest_trinfo().to_state[0].name, 'state2')


class CustomWorkflowTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.member_eid = self.create_user(cnx, 'member').eid

    def test_custom_wf_replace_state_no_history(self):
        """member in inital state with no previous history, state is simply
        redirected when changing workflow
        """
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            wf.add_state('asleep', initial=True)
        with self.admin_access.web_request() as req:
            req.execute('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                     {'wf': wf.eid, 'x': self.member_eid})
            member = req.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.state, 'activated') # no change before commit
            req.cnx.commit()
            member.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_workflow.eid, wf.eid)
            self.assertEqual(iworkflowable.state, 'asleep')
            self.assertEqual(iworkflowable.workflow_history, ())

    def test_custom_wf_replace_state_keep_history(self):
        """member in inital state with some history, state is redirected and
        state change is recorded to history
        """
        with self.admin_access.web_request() as req:
            member = req.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('deactivate')
            iworkflowable.fire_transition('activate')
            req.cnx.commit()
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            wf.add_state('asleep', initial=True)
            shell.rqlexec('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                          {'wf': wf.eid, 'x': self.member_eid})
        with self.admin_access.web_request() as req:
            member = req.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.current_workflow.eid, wf.eid)
            self.assertEqual(iworkflowable.state, 'asleep')
            self.assertEqual(parse_hist(iworkflowable.workflow_history),
                             [('activated', 'deactivated', 'deactivate', None),
                              ('deactivated', 'activated', 'activate', None),
                              ('activated', 'asleep', None, 'workflow changed to "CWUser"')])

    def test_custom_wf_no_initial_state(self):
        """try to set a custom workflow which has no initial state"""
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            wf.add_state('asleep')
            shell.rqlexec('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                          {'wf': wf.eid, 'x': self.member_eid})
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(cm.exception.errors, {'custom_workflow-subject': u'workflow has no initial state'})

    def test_custom_wf_bad_etype(self):
        """try to set a custom workflow which doesn't apply to entity type"""
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'Company')
            wf.add_state('asleep', initial=True)
            shell.rqlexec('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                         {'wf': wf.eid, 'x': self.member_eid})
            with self.assertRaises(ValidationError) as cm:
                shell.commit()
            self.assertEqual(cm.exception.errors, {'custom_workflow-subject': u"workflow isn't a workflow for this type"})

    def test_del_custom_wf(self):
        """member in some state shared by the new workflow, nothing has to be
        done
        """
        with self.admin_access.web_request() as req:
            member = req.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('deactivate')
            req.cnx.commit()
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            wf.add_state('asleep', initial=True)
            shell.rqlexec('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                         {'wf': wf.eid, 'x': self.member_eid})
            shell.commit()
        with self.admin_access.web_request() as req:
            req.execute('DELETE X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                         {'wf': wf.eid, 'x': self.member_eid})
            member = req.entity_from_eid(self.member_eid)
            iworkflowable = member.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.state, 'asleep')# no change before commit
            req.cnx.commit()
            member.cw_clear_all_caches()
            self.assertEqual(iworkflowable.current_workflow.name, "default user workflow")
            self.assertEqual(iworkflowable.state, 'activated')
            self.assertEqual(parse_hist(iworkflowable.workflow_history),
                              [('activated', 'deactivated', 'deactivate', None),
                               ('deactivated', 'asleep', None, 'workflow changed to "CWUser"'),
                               ('asleep', 'activated', None, 'workflow changed to "default user workflow"'),])


class AutoTransitionTC(CubicWebTC):

    def setup_custom_wf(self):
        with self.admin_access.shell() as shell:
            wf = add_wf(shell, 'CWUser')
            asleep = wf.add_state('asleep', initial=True)
            dead = wf.add_state('dead')
            wf.add_transition('rest', asleep, asleep)
            wf.add_transition('sick', asleep, dead, type=u'auto',
                              conditions=({'expr': u'X surname "toto"',
                                           'mainvars': u'X'},))
        return wf

    def test_auto_transition_fired(self):
        wf = self.setup_custom_wf()
        with self.admin_access.web_request() as req:
            user = self.create_user(req, 'member')
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            req.execute('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                        {'wf': wf.eid, 'x': user.eid})
            req.cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, 'asleep')
            self.assertEqual([t.name for t in iworkflowable.possible_transitions()],
                              ['rest'])
            iworkflowable.fire_transition('rest')
            req.cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, 'asleep')
            self.assertEqual([t.name for t in iworkflowable.possible_transitions()],
                              ['rest'])
            self.assertEqual(parse_hist(iworkflowable.workflow_history),
                              [('asleep', 'asleep', 'rest', None)])
            user.cw_set(surname=u'toto') # fulfill condition
            req.cnx.commit()
            iworkflowable.fire_transition('rest')
            req.cnx.commit()
            user.cw_clear_all_caches()
            self.assertEqual(iworkflowable.state, 'dead')
            self.assertEqual(parse_hist(iworkflowable.workflow_history),
                              [('asleep', 'asleep', 'rest', None),
                               ('asleep', 'asleep', 'rest', None),
                               ('asleep', 'dead', 'sick', None),])

    def test_auto_transition_custom_initial_state_fired(self):
        wf = self.setup_custom_wf()
        with self.admin_access.web_request() as req:
            user = self.create_user(req, 'member', surname=u'toto')
            req.execute('SET X custom_workflow WF WHERE X eid %(x)s, WF eid %(wf)s',
                        {'wf': wf.eid, 'x': user.eid})
            req.cnx.commit()
            user.cw_clear_all_caches()
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.state, 'dead')

    def test_auto_transition_initial_state_fired(self):
        with self.admin_access.web_request() as req:
            wf = req.execute('Any WF WHERE ET default_workflow WF, '
                             'ET name %(et)s', {'et': 'CWUser'}).get_entity(0, 0)
            dead = wf.add_state('dead')
            wf.add_transition('sick', wf.state_by_name('activated'), dead,
                              type=u'auto', conditions=({'expr': u'X surname "toto"',
                                                         'mainvars': u'X'},))
            req.cnx.commit()
        with self.admin_access.web_request() as req:
            user = self.create_user(req, 'member', surname=u'toto')
            req.cnx.commit()
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            self.assertEqual(iworkflowable.state, 'dead')


class WorkflowHooksTC(CubicWebTC):

    def setUp(self):
        CubicWebTC.setUp(self)
        with self.admin_access.web_request() as req:
            self.wf = req.user.cw_adapt_to('IWorkflowable').current_workflow
            self.s_activated = self.wf.state_by_name('activated').eid
            self.s_deactivated = self.wf.state_by_name('deactivated').eid
            self.s_dummy = self.wf.add_state(u'dummy').eid
            self.wf.add_transition(u'dummy', (self.s_deactivated,), self.s_dummy)
            ueid = self.create_user(req, 'stduser', commit=False).eid
            # test initial state is set
            rset = req.execute('Any N WHERE S name N, X in_state S, X eid %(x)s',
                                {'x' : ueid})
            self.assertFalse(rset, rset.rows)
            req.cnx.commit()
            initialstate = req.execute('Any N WHERE S name N, X in_state S, X eid %(x)s',
                                        {'x' : ueid})[0][0]
            self.assertEqual(initialstate, u'activated')
            # give access to users group on the user's wf transitions
            # so we can test wf enforcing on euser (managers don't have anymore this
            # enforcement
            req.execute('SET X require_group G '
                         'WHERE G name "users", X transition_of WF, WF eid %(wf)s',
                         {'wf': self.wf.eid})
            req.cnx.commit()

    # XXX currently, we've to rely on hooks to set initial state, or to use execute
    # def test_initial_state(self):
    #     cnx = self.login('stduser')
    #     cu = cnx.cursor()
    #     self.assertRaises(ValidationError, cu.execute,
    #                       'INSERT CWUser X: X login "badaboum", X upassword %(pwd)s, '
    #                       'X in_state S WHERE S name "deactivated"', {'pwd': 'oops'})
    #     cnx.close()
    #     # though managers can do whatever he want
    #     self.execute('INSERT CWUser X: X login "badaboum", X upassword %(pwd)s, '
    #                  'X in_state S, X in_group G WHERE S name "deactivated", G name "users"', {'pwd': 'oops'})
    #     self.commit()

    # test that the workflow is correctly enforced

    def _cleanup_msg(self, msg):
        """remove the variable part of one specific error message"""
        lmsg = msg.split()
        lmsg.pop(1)
        lmsg.pop()
        return ' '.join(lmsg)

    def test_transition_checking1(self):
        with self.new_access('stduser').repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition('activate')
            self.assertEqual(self._cleanup_msg(cm.exception.errors['by_transition-subject']),
                             u"transition isn't allowed from")

    def test_transition_checking2(self):
        with self.new_access('stduser').repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition('dummy')
            self.assertEqual(self._cleanup_msg(cm.exception.errors['by_transition-subject']),
                             u"transition isn't allowed from")

    def test_transition_checking3(self):
        with self.new_access('stduser').repo_cnx() as cnx:
            user = cnx.user
            iworkflowable = user.cw_adapt_to('IWorkflowable')
            iworkflowable.fire_transition('deactivate')
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                iworkflowable.fire_transition('deactivate')
            self.assertEqual(self._cleanup_msg(cm.exception.errors['by_transition-subject']),
                                                u"transition isn't allowed from")
            cnx.rollback()
            # get back now
            iworkflowable.fire_transition('activate')
            cnx.commit()


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

import copy

from cubicweb.devtools import testlib
from cubicweb.wfutils import setup_workflow


class TestWFUtils(testlib.CubicWebTC):

    defs = {
        'group': {
            'etypes': 'CWGroup',
            'default': True,
            'initial_state': u'draft',
            'states': [u'draft', u'published'],
            'transitions': {
                u'publish': {
                    'fromstates': u'draft',
                    'tostate': u'published',
                    'requiredgroups': u'managers'
                }
            }
        }
    }

    def test_create_workflow(self):
        with self.admin_access.cnx() as cnx:
            wf = setup_workflow(cnx, 'group', self.defs['group'])
            self.assertEqual(wf.name, 'group')
            self.assertEqual(wf.initial.name, u'draft')

            draft = wf.state_by_name(u'draft')
            self.assertIsNotNone(draft)

            published = wf.state_by_name(u'published')
            self.assertIsNotNone(published)

            publish = wf.transition_by_name(u'publish')
            self.assertIsNotNone(publish)

            self.assertEqual(publish.destination_state, (published, ))
            self.assertEqual(draft.allowed_transition, (publish, ))

            self.assertEqual(
                {g.name for g in publish.require_group},
                {'managers'})

    def test_update(self):
        with self.admin_access.cnx() as cnx:
            wf = setup_workflow(cnx, 'group', self.defs['group'])
            eid = wf.eid

        with self.admin_access.cnx() as cnx:
            wfdef = copy.deepcopy(self.defs['group'])
            wfdef['states'].append('new')
            wfdef['initial_state'] = 'new'
            wfdef['transitions'][u'publish']['fromstates'] = ('draft', 'new')
            wfdef['transitions'][u'publish']['requiredgroups'] = (
                u'managers', u'users')
            wfdef['transitions'][u'todraft'] = {
                'fromstates': ('new', 'published'),
                'tostate': 'draft',
            }

            wf = setup_workflow(cnx, 'group', wfdef)
            self.assertEqual(wf.eid, eid)
            self.assertEqual(wf.name, 'group')
            self.assertEqual(wf.initial.name, u'new')

            new = wf.state_by_name(u'new')
            self.assertIsNotNone(new)

            draft = wf.state_by_name(u'draft')
            self.assertIsNotNone(draft)

            published = wf.state_by_name(u'published')
            self.assertIsNotNone(published)

            publish = wf.transition_by_name(u'publish')
            self.assertIsNotNone(publish)

            todraft = wf.transition_by_name(u'todraft')
            self.assertIsNotNone(todraft)

            self.assertEqual(
                {g.name for g in publish.require_group},
                {'managers', 'users'})

            self.assertEqual(publish.destination_state, (published, ))
            self.assertEqual(draft.allowed_transition, (publish, ))
            self.assertEqual(todraft.destination_state, (draft, ))
            self.assertEqual(published.allowed_transition, (todraft, ))
            self.assertCountEqual(new.allowed_transition, (publish, todraft))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main, TestCase

from os.path import join

from cubicweb import CW_SOFTWARE_ROOT as BASE, devtools
from cubicweb.cwvreg import CWRegistryStore, UnknownProperty
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.view import EntityAdapter


class YesSchema:
    def __contains__(self, something):
        return True

WEBVIEWSDIR = join(BASE, 'web', 'views')


class VRegistryTC(TestCase):

    def setUp(self):
        config = devtools.TestServerConfiguration('data', __file__)
        self.vreg = CWRegistryStore(config)
        config.bootstrap_cubes()
        self.vreg.schema = config.load_schema()

    def test_load_interface_based_vojects(self):
        self.vreg.init_registration([WEBVIEWSDIR])
        self.vreg.load_file(join(BASE, 'entities', '__init__.py'), 'cubicweb.entities.__init__')
        self.vreg.load_file(join(WEBVIEWSDIR, 'idownloadable.py'), 'cubicweb.web.views.idownloadable')
        self.vreg.load_file(join(WEBVIEWSDIR, 'primary.py'), 'cubicweb.web.views.primary')
        self.assertEqual(len(self.vreg['views']['primary']), 2)
        self.vreg.initialization_completed()
        self.assertEqual(len(self.vreg['views']['primary']), 1)


    def test_load_subinterface_based_appobjects(self):
        self.vreg.register_objects([join(BASE, 'web', 'views', 'idownloadable.py')])
        # check downloadlink was kicked
        self.assertFalse(self.vreg['views'].get('downloadlink'))
        # we've to emulate register_objects to add custom MyCard objects
        path = [join(BASE, 'entities', '__init__.py'),
                join(BASE, 'entities', 'adapters.py'),
                join(BASE, 'web', 'views', 'idownloadable.py')]
        filemods = self.vreg.init_registration(path, None)
        for filepath, modname in filemods:
            self.vreg.load_file(filepath, modname)
        class CardIDownloadableAdapter(EntityAdapter):
            __regid__ = 'IDownloadable'
        self.vreg._loadedmods[__name__] = {}
        self.vreg.register(CardIDownloadableAdapter)
        self.vreg.initialization_completed()
        # check progressbar isn't kicked
        self.assertEqual(len(self.vreg['views']['downloadlink']), 1)

    def test_properties(self):
        self.vreg.reset()
        self.assertNotIn('system.version.cubicweb', self.vreg['propertydefs'])
        self.assertTrue(self.vreg.property_info('system.version.cubicweb'))
        self.assertRaises(UnknownProperty, self.vreg.property_info, 'a.non.existent.key')


class CWVregTC(CubicWebTC):

    def test_property_default_overriding(self):
        # see data/views.py
        from cubicweb.web.views.xmlrss import RSSIconBox
        self.assertEqual(self.vreg.property_info(RSSIconBox._cwpropkey('visible'))['default'], True)

if __name__ == '__main__':
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Core hooks: set generic metadata"""

from datetime import datetime

from pytz import utc

from cubicweb.predicates import is_instance
from cubicweb.server import hook


class MetaDataHook(hook.Hook):
    __abstract__ = True
    category = 'metadata'


class InitMetaAttrsHook(MetaDataHook):
    """before create a new entity -> set creation and modification date

    this is a conveniency hook, you shouldn't have to disable it
    """
    __regid__ = 'metaattrsinit'
    events = ('before_add_entity',)

    def __call__(self):
        timestamp = datetime.now(utc)
        edited = self.entity.cw_edited
        if not edited.get('creation_date'):
            edited['creation_date'] = timestamp
        if not edited.get('modification_date'):
            edited['modification_date'] = timestamp
        if not self._cw.transaction_data.get('do-not-insert-cwuri'):
            cwuri = u'%s%s' % (self._cw.base_url(), self.entity.eid)
            edited.setdefault('cwuri', cwuri)


class UpdateMetaAttrsHook(MetaDataHook):
    """update an entity -> set modification date"""
    __regid__ = 'metaattrsupdate'
    events = ('before_update_entity',)

    def __call__(self):
        # repairing is true during c-c upgrade/shell and similar commands. We
        # usually don't want to update modification date in such cases.
        #
        # XXX to be really clean, we should turn off modification_date update
        # explicitly on each command where we do not want that behaviour.
        if not self._cw.vreg.config.repairing:
            self.entity.cw_edited.setdefault('modification_date', datetime.now(utc))


class SetCreatorOp(hook.DataOperationMixIn, hook.Operation):

    def precommit_event(self):
        cnx = self.cnx
        relations = [(eid, cnx.user.eid) for eid in self.get_data()
                # don't consider entities that have been created and deleted in
                # the same transaction, nor ones where created_by has been
                # explicitly set
                if not cnx.deleted_in_transaction(eid) and \
                   not cnx.entity_from_eid(eid).created_by]
        cnx.add_relations([('created_by', relations)])


class SetOwnershipHook(MetaDataHook):
    """create a new entity -> set owner and creator metadata"""
    __regid__ = 'setowner'
    events = ('after_add_entity',)

    def __call__(self):
        if not self._cw.is_internal_session:
            self._cw.add_relation(self.entity.eid, 'owned_by', self._cw.user.eid)
            SetCreatorOp.get_instance(self._cw).add_data(self.entity.eid)


class SyncOwnersOp(hook.DataOperationMixIn, hook.Operation):
    def precommit_event(self):
        for compositeeid, composedeid in self.get_data():
            if self.cnx.deleted_in_transaction(compositeeid):
                continue
            if self.cnx.deleted_in_transaction(composedeid):
                continue
            self.cnx.execute('SET X owned_by U WHERE C owned_by U, C eid %(c)s,'
                                 'NOT EXISTS(X owned_by U, X eid %(x)s)',
                                 {'c': compositeeid, 'x': composedeid})


class SyncCompositeOwner(MetaDataHook):
    """when adding composite relation, the composed should have the same owners
    has the composite
    """
    __regid__ = 'synccompositeowner'
    events = ('after_add_relation',)

    def __call__(self):
        if self.rtype == 'wf_info_for':
            # skip this special composite relation # XXX (syt) why?
            return
        eidfrom, eidto = self.eidfrom, self.eidto
        composite = self._cw.rtype_eids_rdef(self.rtype, eidfrom, eidto).composite
        if composite == 'subject':
            SyncOwnersOp.get_instance(self._cw).add_data( (eidfrom, eidto) )
        elif composite == 'object':
            SyncOwnersOp.get_instance(self._cw).add_data( (eidto, eidfrom) )


class FixUserOwnershipHook(MetaDataHook):
    """when a user has been created, add owned_by relation on itself"""
    __regid__ = 'fixuserowner'
    __select__ = MetaDataHook.__select__ & is_instance('CWUser')
    events = ('after_add_entity',)

    def __call__(self):
        self._cw.add_relation(self.entity.eid, 'owned_by', self.entity.eid)


class UpdateFTIHook(MetaDataHook):
    """sync fulltext index text index container when a relation with
    fulltext_container set is added / removed
    """
    __regid__ = 'updateftirel'
    events = ('after_add_relation', 'after_delete_relation')

    def __call__(self):
        rtype = self.rtype
        cnx = self._cw
        ftcontainer = cnx.vreg.schema.rschema(rtype).fulltext_container
        if ftcontainer == 'subject':
            cnx.repo.system_source.index_entity(
                cnx, cnx.entity_from_eid(self.eidfrom))
        elif ftcontainer == 'object':
            cnx.repo.system_source.index_entity(
                cnx, cnx.entity_from_eid(self.eidto))



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""
Widgets
~~~~~~~

.. Note::
   A widget is responsible for the display of a field. It may use more than one
   HTML input tags. When the form is posted, a widget is also reponsible to give
   back to the field something it can understand.

   Of course you can not use any widget with any field...

.. autoclass:: cubicweb.web.formwidgets.FieldWidget


HTML <input> based widgets
''''''''''''''''''''''''''

.. autoclass:: cubicweb.web.formwidgets.HiddenInput
.. autoclass:: cubicweb.web.formwidgets.TextInput
.. autoclass:: cubicweb.web.formwidgets.EmailInput
.. autoclass:: cubicweb.web.formwidgets.PasswordSingleInput
.. autoclass:: cubicweb.web.formwidgets.FileInput
.. autoclass:: cubicweb.web.formwidgets.ButtonInput


Other standard HTML widgets
'''''''''''''''''''''''''''

.. autoclass:: cubicweb.web.formwidgets.TextArea
.. autoclass:: cubicweb.web.formwidgets.Select
.. autoclass:: cubicweb.web.formwidgets.CheckBox
.. autoclass:: cubicweb.web.formwidgets.Radio


Date and time widgets
'''''''''''''''''''''

.. autoclass:: cubicweb.web.formwidgets.DateTimePicker
.. autoclass:: cubicweb.web.formwidgets.JQueryDateTimePicker
.. autoclass:: cubicweb.web.formwidgets.JQueryDatePicker
.. autoclass:: cubicweb.web.formwidgets.JQueryTimePicker


Ajax / javascript widgets
'''''''''''''''''''''''''

.. autoclass:: cubicweb.web.formwidgets.FCKEditor
.. autoclass:: cubicweb.web.formwidgets.AjaxWidget
.. autoclass:: cubicweb.web.formwidgets.AutoCompletionWidget
.. autoclass:: cubicweb.web.formwidgets.InOutWidget

.. kill or document StaticFileAutoCompletionWidget
.. kill or document LazyRestrictedAutoCompletionWidget
.. kill or document RestrictedAutoCompletionWidget


Other widgets
'''''''''''''

.. autoclass:: cubicweb.web.formwidgets.PasswordInput
.. autoclass:: cubicweb.web.formwidgets.IntervalWidget
.. autoclass:: cubicweb.web.formwidgets.BitSelect
.. autoclass:: cubicweb.web.formwidgets.HorizontalLayoutWidget
.. autoclass:: cubicweb.web.formwidgets.EditableURLWidget


Form controls
'''''''''''''

Those classes are not proper widget (they are not associated to field) but are
used as form controls. Their API is similar to widgets except that `field`
argument given to :meth:`render` will be `None`.

.. autoclass:: cubicweb.web.formwidgets.Button
.. autoclass:: cubicweb.web.formwidgets.SubmitButton
.. autoclass:: cubicweb.web.formwidgets.ResetButton
.. autoclass:: cubicweb.web.formwidgets.ImgButton
"""


from functools import reduce
from datetime import date

from six import text_type, string_types

from logilab.mtconverter import xml_escape
from logilab.common.date import todatetime

from cubicweb import tags, uilib
from cubicweb.utils import json_dumps
from cubicweb.web import stdmsgs, INTERNAL_FIELD_VALUE, ProcessFormError


class FieldWidget(object):
    """The abstract base class for widgets.

    **Attributes**

    Here are standard attributes of a widget, that may be set on concrete class
    to override default behaviours:

    :attr:`needs_js`
       list of javascript files needed by the widget.

    :attr:`needs_css`
       list of css files needed by the widget.

    :attr:`setdomid`
       flag telling if HTML DOM identifier should be set on input.

    :attr:`suffix`
       string to use a suffix when generating input, to ease usage as a
       sub-widgets (eg widget used by another widget)

    :attr:`vocabulary_widget`
       flag telling if this widget expect a vocabulary

    Also, widget instances takes as first argument a `attrs` dictionary which
    will be stored in the attribute of the same name. It contains HTML
    attributes that should be set in the widget's input tag (though concrete
    classes may ignore it).

    .. currentmodule:: cubicweb.web.formwidgets

    **Form generation methods**

    .. automethod:: render
    .. automethod:: _render
    .. automethod:: values
    .. automethod:: attributes

    **Post handling methods**

    .. automethod:: process_field_data

    """
    needs_js = ()
    needs_css = ()
    setdomid = True
    suffix = None
    # does this widget expect a vocabulary
    vocabulary_widget = False

    def __init__(self, attrs=None, setdomid=None, suffix=None):
        if attrs is None:
            attrs = {}
        self.attrs = attrs
        if setdomid is not None:
            # override class's default value
            self.setdomid = setdomid
        if suffix is not None:
            self.suffix = suffix

    def add_media(self, form):
        """adds media (CSS & JS) required by this widget"""
        if self.needs_js:
            form._cw.add_js(self.needs_js)
        if self.needs_css:
            form._cw.add_css(self.needs_css)

    def render(self, form, field, renderer=None):
        """Called to render the widget for the given `field` in the given
        `form`.  Return a unicode string containing the HTML snippet.

        You will usually prefer to override the :meth:`_render` method so you
        don't have to handle addition of needed javascript / css files.
        """
        self.add_media(form)
        return self._render(form, field, renderer)

    def _render(self, form, field, renderer):
        """This is the method you have to implement in concrete widget classes.
        """
        raise NotImplementedError()

    def format_value(self, form, field, value):
        return field.format_value(form._cw, value)

    def attributes(self, form, field):
        """Return HTML attributes for the widget, automatically setting DOM
        identifier when desired (see :attr:`setdomid` attribute)
        """
        attrs = dict(self.attrs)
        if self.setdomid:
            attrs['id'] = field.dom_id(form, self.suffix)
        if 'placeholder' in attrs:
            attrs['placeholder'] = form._cw._(attrs['placeholder'])
        return attrs

    def values(self, form, field):
        """Return the current *string* values (i.e. for display in an HTML
        string) for the given field. This method returns a list of values since
        it's suitable for all kind of widgets, some of them taking multiple
        values, but you'll get a single value in the list in most cases.

        Those values are searched in:

        1. previously submitted form values if any (on validation error)

        2. req.form (specified using request parameters)

        3. extra form values given to form.render call (specified the code
           generating the form)

        4. field's typed value (returned by its
           :meth:`~cubicweb.web.formfields.Field.typed_value` method)

        Values found in 1. and 2. are expected te be already some 'display
        value' (eg a string) while those found in 3. and 4. are expected to be
        correctly typed value.

        3 and 4 are handle by the :meth:`typed_value` method to ease reuse in
        concrete classes.
        """
        values = None
        if not field.ignore_req_params:
            qname = field.input_name(form, self.suffix)
            # value from a previous post that has raised a validation error
            if qname in form.form_previous_values:
                values = form.form_previous_values[qname]
            # value specified using form parameters
            elif qname in form._cw.form:
                values = form._cw.form[qname]
            elif field.name != qname and field.name in form._cw.form:
                # XXX compat: accept attr=value in req.form to specify value of
                # attr-subject
                values = form._cw.form[field.name]
        if values is None:
            values = self.typed_value(form, field)
            if values != INTERNAL_FIELD_VALUE:
                values = self.format_value(form, field, values)
        if not isinstance(values, (tuple, list)):
            values = (values,)
        return values

    def typed_value(self, form, field):
        """return field's *typed* value specified in:
        3. extra form values given to render()
        4. field's typed value
        """
        qname = field.input_name(form)
        for key in ((field, form), qname):
            try:
                return form.formvalues[key]
            except KeyError:
                continue
        if field.name != qname and field.name in form.formvalues:
            return form.formvalues[field.name]
        return field.typed_value(form)

    def process_field_data(self, form, field):
        """Return process posted value(s) for widget and return something
        understandable by the associated `field`. That value may be correctly
        typed or a string that the field may parse.
        """
        posted = form._cw.form
        val = posted.get(field.input_name(form, self.suffix))
        if isinstance(val, string_types):
            val = val.strip()
        return val

    # XXX deprecates
    def values_and_attributes(self, form, field):
        return self.values(form, field), self.attributes(form, field)


class Input(FieldWidget):
    """abstract widget class for <input> tag based widgets"""
    type = None

    def _render(self, form, field, renderer):
        """render the widget for the given `field` of `form`.

        Generate one <input> tag for each field's value
        """
        values, attrs = self.values_and_attributes(form, field)
        # ensure something is rendered
        if not values:
            values = (INTERNAL_FIELD_VALUE,)
        inputs = [tags.input(name=field.input_name(form, self.suffix),
                             type=self.type, value=value, **attrs)
                  for value in values]
        return u'\n'.join(inputs)


# basic html widgets ###########################################################

class TextInput(Input):
    """Simple <input type='text'>, will return a unicode string."""
    type = 'text'


class EmailInput(Input):
    """Simple <input type='email'>, will return a unicode string."""
    type = 'email'


class PasswordSingleInput(Input):
    """Simple <input type='password'>, will return a utf-8 encoded string.

    You may prefer using the :class:`~cubicweb.web.formwidgets.PasswordInput`
    widget which handles password confirmation.
    """
    type = 'password'

    def process_field_data(self, form, field):
        value = super(PasswordSingleInput, self).process_field_data(form, field)
        if value is not None:
            return value.encode('utf-8')
        return value


class PasswordInput(Input):
    """<input type='password'> and a confirmation input. Form processing will
    fail if password and confirmation differs, else it will return the password
    as a utf-8 encoded string.
    """
    type = 'password'

    def _render(self, form, field, renderer):
        assert self.suffix is None, 'suffix not supported'
        values, attrs = self.values_and_attributes(form, field)
        assert len(values) == 1
        domid = attrs.pop('id')
        inputs = [tags.input(name=field.input_name(form),
                             value=values[0], type=self.type, id=domid, **attrs),
                  '<br/>',
                  tags.input(name=field.input_name(form, '-confirm'),
                             value=values[0], type=self.type, **attrs),
                  '&#160;', tags.span(form._cw._('confirm password'),
                                      **{'class': 'emphasis'})]
        return u'\n'.join(inputs)

    def process_field_data(self, form, field):
        passwd1 = super(PasswordInput, self).process_field_data(form, field)
        passwd2 = form._cw.form.get(field.input_name(form, '-confirm'))
        if passwd1 == passwd2:
            if passwd1 is None:
                return None
            return passwd1.encode('utf-8')
        raise ProcessFormError(form._cw._("password and confirmation don't match"))


class FileInput(Input):
    """Simple <input type='file'>, will return a tuple (name, stream) where
    name is the posted file name and stream a file like object containing the
    posted file data.
    """
    type = 'file'

    def values(self, form, field):
        # ignore value which makes no sense here (XXX even on form validation error?)
        return ('',)


class HiddenInput(Input):
    """Simple <input type='hidden'> for hidden value, will return a unicode
    string.
    """
    type = 'hidden'
    setdomid = False  # by default, don't set id attribute on hidden input


class ButtonInput(Input):
    """Simple <input type='button'>, will return a unicode string.

    If you want a global form button, look at the :class:`Button`,
    :class:`SubmitButton`, :class:`ResetButton` and :class:`ImgButton` below.
    """
    type = 'button'


class TextArea(FieldWidget):
    """Simple <textarea>, will return a unicode string."""
    _minrows = 2
    _maxrows = 15
    _columns = 80

    def _render(self, form, field, renderer):
        values, attrs = self.values_and_attributes(form, field)
        attrs.setdefault('onkeyup', 'autogrow(this)')
        if not values:
            value = u''
        elif len(values) == 1:
            value = values[0]
        else:
            raise ValueError('a textarea is not supposed to be multivalued')
        lines = value.splitlines()
        linecount = len(lines)
        for line in lines:
            linecount += len(line) // self._columns
        attrs.setdefault('cols', self._columns)
        attrs.setdefault('rows', min(self._maxrows, linecount + self._minrows))
        return tags.textarea(value, name=field.input_name(form, self.suffix),
                             **attrs)


class FCKEditor(TextArea):
    """FCKEditor enabled <textarea>, will return a unicode string containing
    HTML formated text.
    """
    def __init__(self, *args, **kwargs):
        super(FCKEditor, self).__init__(*args, **kwargs)
        self.attrs['cubicweb:type'] = 'wysiwyg'

    def _render(self, form, field, renderer):
        form._cw.fckeditor_config()
        return super(FCKEditor, self)._render(form, field, renderer)


class Select(FieldWidget):
    """Simple <select>, for field having a specific vocabulary. Will return
    a unicode string, or a list of unicode strings.
    """
    vocabulary_widget = True
    default_size = 10

    def __init__(self, attrs=None, multiple=False, **kwargs):
        super(Select, self).__init__(attrs, **kwargs)
        self._multiple = multiple

    def _render(self, form, field, renderer):
        curvalues, attrs = self.values_and_attributes(form, field)
        options = []
        optgroup_opened = False
        vocab = field.vocabulary(form)
        for option in vocab:
            try:
                label, value, oattrs = option
            except ValueError:
                label, value = option
                oattrs = {}
            if value is None:
                # handle separator
                if optgroup_opened:
                    options.append(u'</optgroup>')
                oattrs.setdefault('label', label or '')
                options.append(u'<optgroup %s>' % uilib.sgml_attributes(oattrs))
                optgroup_opened = True
            elif self.value_selected(value, curvalues):
                options.append(tags.option(label, value=value,
                                           selected='selected', **oattrs))
            else:
                options.append(tags.option(label, value=value, **oattrs))
        if optgroup_opened:
            options.append(u'</optgroup>')
        if 'size' not in attrs:
            if self._multiple:
                size = text_type(min(self.default_size, len(vocab) or 1))
            else:
                size = u'1'
            attrs['size'] = size
        return tags.select(name=field.input_name(form, self.suffix),
                           multiple=self._multiple, options=options, **attrs)

    def value_selected(self, value, curvalues):
        return value in curvalues


class InOutWidget(Select):
    needs_js = ('cubicweb.widgets.js', )
    default_size = 10
    template = """
<table id="%(widgetid)s">
  <tr>
    <td>%(inoutinput)s</td>
    <td><div style="margin-bottom:3px">%(addinput)s</div>
        <div>%(removeinput)s</div>
    </td>
    <td>%(resinput)s</td>
  </tr>
</table>
"""
    add_button = ('<input type="button" class="wdgButton cwinoutadd" '
                  'value="&gt;&gt;" size="10" />')
    remove_button = ('<input type="button" class="wdgButton cwinoutremove" '
                     'value="&lt;&lt;" size="10" />')

    def __init__(self, *args, **kwargs):
        super(InOutWidget, self).__init__(*args, **kwargs)
        self._multiple = True

    def render_select(self, form, field, name, selected=False):
        values, attrs = self.values_and_attributes(form, field)
        options = []
        inputs = []
        for option in field.vocabulary(form):
            try:
                label, value, _oattrs = option
            except ValueError:
                label, value = option
            if selected:
                # add values
                if value in values:
                    options.append(tags.option(label, value=value))
                    # add hidden inputs
                    inputs.append(tags.input(value=value,
                                             name=field.dom_id(form),
                                             type="hidden"))
            else:
                if value not in values:
                    options.append(tags.option(label, value=value))
        if 'size' not in attrs:
            attrs['size'] = self.default_size
        if 'id' in attrs:
            attrs.pop('id')
        return tags.select(name=name, multiple=self._multiple, id=name,
                           options=options, **attrs) + '\n'.join(inputs)

    def _render(self, form, field, renderer):
        domid = field.dom_id(form)
        jsnodes = {'widgetid': domid,
                   'from': 'from_' + domid,
                   'to': 'to_' + domid}
        form._cw.add_onload(u'$(cw.jqNode("%s")).cwinoutwidget("%s", "%s");'
                            % (jsnodes['widgetid'], jsnodes['from'], jsnodes['to']))
        field.required = True
        return (self.template %
                {'widgetid': jsnodes['widgetid'],
                 # helpinfo select tag
                 'inoutinput': self.render_select(form, field, jsnodes['from']),
                 # select tag with resultats
                 'resinput': self.render_select(form, field, jsnodes['to'], selected=True),
                 'addinput': self.add_button % jsnodes,
                 'removeinput': self.remove_button % jsnodes
                 })


class BitSelect(Select):
    """Select widget for IntField using a vocabulary with bit masks as values.

    See also :class:`~cubicweb.web.facet.BitFieldFacet`.
    """
    def __init__(self, attrs=None, multiple=True, **kwargs):
        super(BitSelect, self).__init__(attrs, multiple=multiple, **kwargs)

    def value_selected(self, value, curvalues):
        mask = reduce(lambda x, y: int(x) | int(y), curvalues, 0)
        return int(value) & mask

    def process_field_data(self, form, field):
        """Return process posted value(s) for widget and return something
        understandable by the associated `field`. That value may be correctly
        typed or a string that the field may parse.
        """
        val = super(BitSelect, self).process_field_data(form, field)
        if isinstance(val, list):
            val = reduce(lambda x, y: int(x) | int(y), val, 0)
        elif val:
            val = int(val)
        else:
            val = 0
        return val


class CheckBox(Input):
    """Simple <input type='checkbox'>, for field having a specific
    vocabulary. One input will be generated for each possible value.

    You can specify separator using the `separator` constructor argument, by
    default <br/> is used.
    """
    type = 'checkbox'
    default_separator = u'<br/>\n'
    vocabulary_widget = True

    def __init__(self, attrs=None, separator=None, **kwargs):
        super(CheckBox, self).__init__(attrs, **kwargs)
        self.separator = separator or self.default_separator

    def _render(self, form, field, renderer):
        curvalues, attrs = self.values_and_attributes(form, field)
        domid = attrs.pop('id', None)
        sep = self.separator
        options = []
        for i, option in enumerate(field.vocabulary(form)):
            try:
                label, value, oattrs = option
            except ValueError:
                label, value = option
                oattrs = {}
            iattrs = attrs.copy()
            iattrs.update(oattrs)
            if i == 0 and domid is not None:
                iattrs.setdefault('id', domid)
            if value in curvalues:
                iattrs['checked'] = u'checked'
            tag = tags.input(name=field.input_name(form, self.suffix),
                             type=self.type, value=value, **iattrs)
            options.append(u'<label>%s&#160;%s</label>' % (tag, xml_escape(label)))
        return sep.join(options)


class Radio(CheckBox):
    """Simle <input type='radio'>, for field having a specific vocabulary. One
    input will be generated for each possible value.

    You can specify separator using the `separator` constructor argument, by
    default <br/> is used.
    """
    type = 'radio'


# javascript widgets ###########################################################

class DateTimePicker(TextInput):
    """<input type='text'> + javascript date/time picker for date or datetime
    fields. Will return the date or datetime as a unicode string.
    """
    monthnames = ('january', 'february', 'march', 'april',
                  'may', 'june', 'july', 'august',
                  'september', 'october', 'november', 'december')
    daynames = ('monday', 'tuesday', 'wednesday', 'thursday',
                'friday', 'saturday', 'sunday')

    needs_js = ('cubicweb.calendar.js',)
    needs_css = ('cubicweb.calendar_popup.css',)

    @classmethod
    def add_localized_infos(cls, req):
        """inserts JS variables defining localized months and days"""
        _ = req._
        monthnames = [_(mname) for mname in cls.monthnames]
        daynames = [_(dname) for dname in cls.daynames]
        req.html_headers.define_var('MONTHNAMES', monthnames)
        req.html_headers.define_var('DAYNAMES', daynames)

    def _render(self, form, field, renderer):
        txtwidget = super(DateTimePicker, self)._render(form, field, renderer)
        self.add_localized_infos(form._cw)
        cal_button = self._render_calendar_popup(form, field)
        return txtwidget + cal_button

    def _render_calendar_popup(self, form, field):
        value = field.typed_value(form)
        if not value:
            value = date.today()
        inputid = field.dom_id(form)
        helperid = '%shelper' % inputid
        year, month = value.year, value.month
        return (u"""<a onclick="toggleCalendar('%s', '%s', %s, %s);" class="calhelper">
<img src="%s" title="%s" alt="" /></a><div class="calpopup hidden" id="%s"></div>"""
                % (helperid, inputid, year, month,
                   form._cw.uiprops['CALENDAR_ICON'],
                   form._cw._('calendar'), helperid))


class JQueryDatePicker(FieldWidget):
    """Use jquery.ui.datepicker to define a date picker. Will return the date as
    a unicode string.

    You can couple DatePickers by using the min_of and/or max_of parameters.
    The DatePicker identified by the value of min_of(/max_of) will force the user to
    choose a date anterior(/posterior) to this DatePicker.

    example:

    start and end are two JQueryDatePicker and start must always be before end::

        affk.set_field_kwargs(etype, 'start_date', widget=JQueryDatePicker(min_of='end_date'))
        affk.set_field_kwargs(etype, 'end_date', widget=JQueryDatePicker(max_of='start_date'))

    That way, on change of end(/start) value a new max(/min) will be set for start(/end)
    The invalid dates will be gray colored in the datepicker
    """
    needs_js = ('jquery.ui.js', )
    needs_css = ('jquery.ui.css',)
    default_size = 10

    def __init__(self, datestr=None, min_of=None, max_of=None, **kwargs):
        super(JQueryDatePicker, self).__init__(**kwargs)
        self.min_of = min_of
        self.max_of = max_of
        self.value = datestr

    def attributes(self, form, field):
        form._cw.add_js('cubicweb.widgets.js')
        attrs = super(JQueryDatePicker, self).attributes(form, field)
        if self.max_of:
            attrs['data-max-of'] = '%s-subject:%s' % (self.max_of, form.edited_entity.eid)
        if self.min_of:
            attrs['data-min-of'] = '%s-subject:%s' % (self.min_of, form.edited_entity.eid)
        return attrs

    def _render(self, form, field, renderer):
        req = form._cw
        if req.lang != 'en':
            req.add_js('jquery.ui.datepicker-%s.js' % req.lang)
        domid = field.dom_id(form, self.suffix)
        # XXX find a way to understand every format
        fmt = req.property_value('ui.date-format')
        picker_fmt = fmt.replace('%Y', 'yy').replace('%m', 'mm').replace('%d', 'dd')
        max_date = min_date = None
        if self.min_of:
            current = getattr(form.edited_entity, self.min_of)
            if current is not None:
                max_date = current.strftime(fmt)
        if self.max_of:
            current = getattr(form.edited_entity, self.max_of)
            if current is not None:
                min_date = current.strftime(fmt)
        req.add_onload(u'renderJQueryDatePicker("%s", "%s", "%s", %s, %s);'
                       % (domid, req.uiprops['CALENDAR_ICON'], picker_fmt, json_dumps(min_date),
                          json_dumps(max_date)))
        return self._render_input(form, field)

    def _render_input(self, form, field):
        if self.value is None:
            value = self.values(form, field)[0]
        else:
            value = self.value
        attrs = self.attributes(form, field)
        attrs.setdefault('size', text_type(self.default_size))
        return tags.input(name=field.input_name(form, self.suffix),
                          value=value, type='text', **attrs)


class JQueryTimePicker(JQueryDatePicker):
    """Use jquery.timePicker to define a time picker. Will return the time as a
    unicode string.
    """
    needs_js = ('jquery.timePicker.js',)
    needs_css = ('jquery.timepicker.css',)
    default_size = 5

    def __init__(self, timestr=None, timesteps=30, separator=u':', **kwargs):
        super(JQueryTimePicker, self).__init__(timestr, **kwargs)
        self.timesteps = timesteps
        self.separator = separator

    def _render(self, form, field, renderer):
        domid = field.dom_id(form, self.suffix)
        form._cw.add_onload(u'cw.jqNode("%s").timePicker({step: %s, separator: "%s"})' % (
            domid, self.timesteps, self.separator))
        return self._render_input(form, field)


class JQueryDateTimePicker(FieldWidget):
    """Compound widget using :class:`JQueryDatePicker` and
    :class:`JQueryTimePicker` widgets to define a date and time picker. Will
    return the date and time as python datetime instance.
    """
    def __init__(self, initialtime=None, timesteps=15, **kwargs):
        super(JQueryDateTimePicker, self).__init__(**kwargs)
        self.initialtime = initialtime
        self.timesteps = timesteps

    def _render(self, form, field, renderer):
        """render the widget for the given `field` of `form`.

        Generate one <input> tag for each field's value
        """
        req = form._cw
        dateqname = field.input_name(form, 'date')
        timeqname = field.input_name(form, 'time')
        if dateqname in form.form_previous_values:
            datestr = form.form_previous_values[dateqname]
            timestr = form.form_previous_values[timeqname]
        else:
            datestr = timestr = u''
            if field.name in req.form:
                value = req.parse_datetime(req.form[field.name])
            else:
                value = self.typed_value(form, field)
            if value:
                datestr = req.format_date(value)
                timestr = req.format_time(value)
            elif self.initialtime:
                timestr = req.format_time(self.initialtime)
        datepicker = JQueryDatePicker(datestr=datestr, suffix='date')
        timepicker = JQueryTimePicker(timestr=timestr, timesteps=self.timesteps,
                                      suffix='time')
        return u'<div id="%s">%s%s</div>' % (field.dom_id(form),
                                             datepicker.render(form, field, renderer),
                                             timepicker.render(form, field, renderer))

    def process_field_data(self, form, field):
        req = form._cw
        datestr = req.form.get(field.input_name(form, 'date'))
        if datestr:
            datestr = datestr.strip()
        if not datestr:
            return None
        try:
            date = todatetime(req.parse_datetime(datestr, 'Date'))
        except ValueError as exc:
            raise ProcessFormError(text_type(exc))
        timestr = req.form.get(field.input_name(form, 'time'))
        if timestr:
            timestr = timestr.strip()
        if not timestr:
            return date
        try:
            time = req.parse_datetime(timestr, 'Time')
        except ValueError as exc:
            raise ProcessFormError(text_type(exc))
        return date.replace(hour=time.hour, minute=time.minute, second=time.second)


# ajax widgets ################################################################

def init_ajax_attributes(attrs, wdgtype, loadtype=u'auto'):
    try:
        attrs['class'] += u' widget'
    except KeyError:
        attrs['class'] = u'widget'
    attrs.setdefault('cubicweb:wdgtype', wdgtype)
    attrs.setdefault('cubicweb:loadtype', loadtype)


class AjaxWidget(FieldWidget):
    """Simple <div> based ajax widget, requiring a `wdgtype` argument telling
    which javascript widget should be used.
    """
    def __init__(self, wdgtype, inputid=None, **kwargs):
        super(AjaxWidget, self).__init__(**kwargs)
        init_ajax_attributes(self.attrs, wdgtype)
        if inputid is not None:
            self.attrs['cubicweb:inputid'] = inputid

    def _render(self, form, field, renderer):
        attrs = self.values_and_attributes(form, field)[-1]
        return tags.div(**attrs)


class AutoCompletionWidget(TextInput):
    """<input type='text'> based ajax widget, taking a `autocomplete_initfunc`
    argument which should specify the name of a method of the json
    controller. This method is expected to return allowed values for the input,
    that the widget will use to propose matching values as you type.
    """
    needs_js = ('cubicweb.widgets.js', 'jquery.ui.js')
    needs_css = ('jquery.ui.css',)
    default_settings = {}

    def __init__(self, *args, **kwargs):
        self.autocomplete_settings = kwargs.pop('autocomplete_settings',
                                                self.default_settings)
        self.autocomplete_initfunc = kwargs.pop('autocomplete_initfunc')
        super(AutoCompletionWidget, self).__init__(*args, **kwargs)

    def values(self, form, field):
        values = super(AutoCompletionWidget, self).values(form, field)
        if not values:
            values = ('',)
        return values

    def _render(self, form, field, renderer):
        entity = form.edited_entity
        domid = field.dom_id(form).replace(':', r'\\:')
        if callable(self.autocomplete_initfunc):
            data = self.autocomplete_initfunc(form, field)
        else:
            data = xml_escape(self._get_url(entity, field))
        form._cw.add_onload(u'$("#%s").cwautocomplete(%s, %s);'
                            % (domid, json_dumps(data),
                               json_dumps(self.autocomplete_settings)))
        return super(AutoCompletionWidget, self)._render(form, field, renderer)

    def _get_url(self, entity, field):
        fname = self.autocomplete_initfunc
        return entity._cw.build_url('ajax', fname=fname, mode='remote',
                                    pageid=entity._cw.pageid)


class StaticFileAutoCompletionWidget(AutoCompletionWidget):
    """XXX describe me"""
    wdgtype = 'StaticFileSuggestField'

    def _get_url(self, entity, field):
        return entity._cw.data_url(self.autocomplete_initfunc)


class RestrictedAutoCompletionWidget(AutoCompletionWidget):
    """XXX describe me"""
    default_settings = {'mustMatch': True}


class LazyRestrictedAutoCompletionWidget(RestrictedAutoCompletionWidget):
    """remote autocomplete """

    def values_and_attributes(self, form, field):
        """override values_and_attributes to handle initial displayed values"""
        values, attrs = super(LazyRestrictedAutoCompletionWidget, self).values_and_attributes(
            form, field)
        assert len(values) == 1, "multiple selection is not supported yet by LazyWidget"
        if not values[0]:
            values = form.cw_extra_kwargs.get(field.name, '')
            if not isinstance(values, (tuple, list)):
                values = (values,)
        try:
            values = list(values)
            values[0] = int(values[0])
            attrs['cubicweb:initialvalue'] = values[0]
            values = (self.display_value_for(form, values[0]),)
        except (TypeError, ValueError):
            pass
        return values, attrs

    def display_value_for(self, form, value):
        entity = form._cw.entity_from_eid(value)
        return entity.view('combobox')


# more widgets #################################################################

class IntervalWidget(FieldWidget):
    """Custom widget to display an interval composed by 2 fields. This widget is
    expected to be used with a :class:`CompoundField` containing the two actual
    fields.

    Exemple usage::

      class MyForm(FieldsForm):
         price = CompoundField(fields=(IntField(name='minprice'),
                                       IntField(name='maxprice')),
                               label=_('price'),
                               widget=IntervalWidget())
    """
    def _render(self, form, field, renderer):
        actual_fields = field.fields
        assert len(actual_fields) == 2
        return u'<div>%s %s %s %s</div>' % (
            form._cw._('from_interval_start'),
            actual_fields[0].render(form, renderer),
            form._cw._('to_interval_end'),
            actual_fields[1].render(form, renderer),
        )


class HorizontalLayoutWidget(FieldWidget):
    """Custom widget to display a set of fields grouped together horizontally in
    a form. See `IntervalWidget` for example usage.
    """
    def _render(self, form, field, renderer):
        if self.attrs.get('display_label', True):
            subst = self.attrs.get('label_input_substitution', '%(label)s %(input)s')
            fields = [subst % {'label': renderer.render_label(form, f),
                               'input': f.render(form, renderer)}
                      for f in field.subfields(form)]
        else:
            fields = [f.render(form, renderer) for f in field.subfields(form)]
        return u'<div>%s</div>' % ' '.join(fields)


class EditableURLWidget(FieldWidget):
    """Custom widget to edit separatly a URL path / query string (used by
    default for the `path` attribute of `Bookmark` entities).

    It deals with url quoting nicely so that the user edit the unquoted value.
    """

    def _render(self, form, field, renderer):
        assert self.suffix is None, 'not supported'
        req = form._cw
        pathqname = field.input_name(form, 'path')
        fqsqname = field.input_name(form, 'fqs')  # formatted query string
        if pathqname in form.form_previous_values:
            path = form.form_previous_values[pathqname]
            fqs = form.form_previous_values[fqsqname]
        else:
            if field.name in req.form:
                value = req.form[field.name]
            else:
                value = self.typed_value(form, field)
            if value:
                try:
                    path, qs = value.split('?', 1)
                except ValueError:
                    path = value
                    qs = ''
            else:
                path = qs = ''
            fqs = u'\n'.join(u'%s=%s' % (k, v) for k, v in req.url_parse_qsl(qs))
        attrs = dict(self.attrs)
        if self.setdomid:
            attrs['id'] = field.dom_id(form)
        # ensure something is rendered
        inputs = [u'<table><tr><th>',
                  req._('i18n_bookmark_url_path'),
                  u'</th><td>',
                  tags.input(name=pathqname, type='string', value=path, **attrs),
                  u'</td></tr><tr><th>',
                  req._('i18n_bookmark_url_fqs'),
                  u'</th><td>']
        if self.setdomid:
            attrs['id'] = field.dom_id(form, 'fqs')
        attrs.setdefault('cols', 60)
        attrs.setdefault('onkeyup', 'autogrow(this)')
        inputs += [tags.textarea(fqs, name=fqsqname, **attrs),
                   u'</td></tr></table>']
        # surrounding div necessary for proper error localization
        return u'<div id="%s">%s</div>' % (
            field.dom_id(form), u'\n'.join(inputs))

    def process_field_data(self, form, field):
        req = form._cw
        values = {}
        path = req.form.get(field.input_name(form, 'path'))
        if isinstance(path, string_types):
            path = path.strip()
        if path is None:
            path = u''
        fqs = req.form.get(field.input_name(form, 'fqs'))
        if isinstance(fqs, string_types):
            fqs = fqs.strip() or None
            if fqs:
                for i, line in enumerate(fqs.split('\n')):
                    line = line.strip()
                    if line:
                        try:
                            key, val = line.split('=', 1)
                        except ValueError:
                            msg = req._("wrong query parameter line %s") % (i + 1)
                            raise ProcessFormError(msg)
                        # value will be url quoted by build_url_params
                        values.setdefault(key, []).append(val)
        if not values:
            return path
        return u'%s?%s' % (path, req.build_url_params(**values))


# form controls ######################################################################

class Button(Input):
    """Simple <input type='button'>, base class for global form buttons.

    Note that `label` is a msgid which will be translated at form generation
    time, you should not give an already translated string.
    """
    type = 'button'
    css_class = 'validateButton'

    def __init__(self, label=stdmsgs.BUTTON_OK, attrs=None,
                 setdomid=None,
                 name='', value='', onclick=None, cwaction=None):
        super(Button, self).__init__(attrs, setdomid)
        if isinstance(label, tuple):
            self.label = label[0]
            self.icon = label[1]
        else:
            self.label = label
            self.icon = None
        self.name = name
        self.value = ''
        self.onclick = onclick
        self.cwaction = cwaction

    def render(self, form, field=None, renderer=None):
        label = form._cw._(self.label)
        attrs = self.attrs.copy()
        attrs.setdefault('class', self.css_class)
        if self.cwaction:
            assert self.onclick is None
            attrs['onclick'] = "postForm('__action_%s', \'%s\', \'%s\')" % (
                self.cwaction, self.label, form.domid)
        elif self.onclick:
            attrs['onclick'] = self.onclick
        if self.name:
            attrs['name'] = self.name
            if self.setdomid:
                attrs['id'] = self.name
        if self.icon:
            img = tags.img(src=form._cw.uiprops[self.icon], alt=self.icon)
        else:
            img = u''
        return tags.button(img + xml_escape(label), escapecontent=False,
                           value=label, type=self.type, **attrs)


class SubmitButton(Button):
    """Simple <input type='submit'>, main button to submit a form"""
    type = 'submit'


class ResetButton(Button):
    """Simple <input type='reset'>, main button to reset a form. You usually
    don't want to use this.
    """
    type = 'reset'


class ImgButton(object):
    """Simple <img> wrapped into a <a> tag with href triggering something (usually a
    javascript call).
    """
    def __init__(self, domid, href, label, imgressource):
        self.domid = domid
        self.href = href
        self.imgressource = imgressource
        self.label = label

    def render(self, form, field=None, renderer=None):
        label = form._cw._(self.label)
        imgsrc = form._cw.uiprops[self.imgressource]
        return '<a id="%(domid)s" href="%(href)s">'\
               '<img src="%(imgsrc)s" alt="%(label)s"/>%(label)s</a>' % {
                   'label': label, 'imgsrc': imgsrc,
                   'domid': self.domid, 'href': self.href}



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""json export views"""


from cubicweb import _

from cubicweb.uilib import rest_traceback

from cubicweb.utils import json_dumps
from cubicweb.predicates import ExpectedValuePredicate, any_rset, empty_rset
from cubicweb.view import EntityView, AnyRsetView
from cubicweb.web.application import anonymized_request
from cubicweb.web.views import basecontrollers, management


class JsonpController(basecontrollers.ViewController):
    """The jsonp controller is the same as a ViewController but :

    - anonymize request (avoid CSRF attacks)
    - if ``vid`` parameter is passed, make sure it's sensible (i.e. either
      "jsonexport" or "ejsonexport")
    - if ``callback`` request parameter is passed, it's used as json padding


    Response's content-type will either be ``application/javascript`` or
    ``application/json`` depending on ``callback`` parameter presence or not.
    """
    __regid__ = 'jsonp'

    def publish(self, rset=None):
        if 'vid' in self._cw.form:
            vid = self._cw.form['vid']
            if vid not in ('jsonexport', 'ejsonexport'):
                self.warning("vid %s can't be used with jsonp controller, "
                             "falling back to jsonexport", vid)
                self._cw.form['vid'] = 'jsonexport'
        else:  # if no vid is specified, use jsonexport
            self._cw.form['vid'] = 'jsonexport'
        if self._cw.vreg.config['anonymize-jsonp-queries']:
            with anonymized_request(self._cw):
                return self._get_json_data(rset)
        else:
            return self._get_json_data(rset)

    def _get_json_data(self, rset):
        json_data = super(JsonpController, self).publish(rset)
        if 'callback' in self._cw.form:  # jsonp
            json_padding = self._cw.form['callback'].encode('ascii')
            # use ``application/javascript`` if ``callback`` parameter is
            # provided, keep ``application/json`` otherwise
            self._cw.set_content_type('application/javascript')
            json_data = json_padding + b'(' + json_data + b')'
        return json_data


class JsonMixIn(object):
    """mixin class for json views

    Handles the following optional request parameters:

    - ``_indent`` : must be an integer. If found, it is used to pretty print
      json output
    """
    templatable = False
    content_type = 'application/json'
    binary = True

    def wdata(self, data):
        if '_indent' in self._cw.form:
            indent = int(self._cw.form['_indent'])
        else:
            indent = None
        # python's json.dumps escapes non-ascii characters
        self.w(json_dumps(data, indent=indent).encode('ascii'))


class JsonRsetView(JsonMixIn, AnyRsetView):
    """dumps raw result set in JSON format"""
    __regid__ = 'jsonexport'
    __select__ = any_rset()  # means rset might be empty or have any shape
    title = _('json-export-view')

    def call(self):
        # XXX mimic w3c recommandations to serialize SPARQL results in json?
        #     http://www.w3.org/TR/rdf-sparql-json-res/
        self.wdata(self.cw_rset.rows)


class JsonEntityView(JsonMixIn, EntityView):
    """dumps rset entities in JSON

    The following additional metadata is added to each row :

    - ``cw_etype`` : entity type
    - ``cw_source`` : source url
    """
    __regid__ = 'ejsonexport'
    __select__ = EntityView.__select__ | empty_rset()
    title = _('json-entities-export-view')

    def call(self):
        entities = []
        for entity in self.cw_rset.entities():
            serializer = entity.cw_adapt_to('ISerializable')
            entities.append(serializer.serialize())
        self.wdata(entities)


class _requested_vid(ExpectedValuePredicate):
    """predicate that checks vid parameter value

    It differs from ``match_view`` in that it doesn't expect a ``view``
    parameter to be given to ``select`` but will rather check
    ``req.form['vid']`` to match expected vid.
    """
    def __call__(self, cls, req, rset=None, **kwargs):
        return req.form.get('vid') in self.expected


class JsonErrorView(JsonMixIn, management.ErrorView):
    """custom error view selected when client asks for a json view

    The returned json object will contain err / traceback informations.
    """
    __select__ = (management.ErrorView.__select__ &
                  _requested_vid('jsonexport', 'ejsonexport'))

    def call(self):
        errmsg, exclass, excinfo = self._excinfo()
        self.wdata({
            'errmsg': errmsg,
            'exclass': exclass,
            'traceback': rest_traceback(excinfo, errmsg),
        })



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""some hooks to handle notification on entity's changes"""



from logilab.common.textutils import normalize_text
from logilab.common.deprecation import deprecated

from cubicweb import RegistryNotFound
from cubicweb.predicates import is_instance
from cubicweb.server import hook
from cubicweb.sobjects.supervising import SupervisionMailOp


@deprecated('[3.17] use notify_on_commit instead')
def RenderAndSendNotificationView(cnx, view, viewargs=None):
    notify_on_commit(cnx, view, viewargs)


def notify_on_commit(cnx, view, viewargs=None):
    """register a notification view (see
    :class:`~cubicweb.sobjects.notification.NotificationView`) to be sent at
    post-commit time, ie only if the transaction has succeeded.

    `viewargs` is an optional dictionary containing extra argument to be given
    to :meth:`~cubicweb.sobjects.notification.NotificationView.render_and_send`
    """
    if viewargs is None:
        viewargs = {}
    notif_op = _RenderAndSendNotificationOp.get_instance(cnx)
    notif_op.add_data((view, viewargs))


class _RenderAndSendNotificationOp(hook.DataOperationMixIn, hook.Operation):
    """End of the notification chain. Do render and send views after commit

    All others Operations end up adding data to this Operation.
    The notification are done on ``postcommit_event`` to make sure to prevent
    sending notification about rolled back data.
    """

    containercls = list

    def postcommit_event(self):
        deleted = self.cnx.deleted_in_transaction
        for view, viewargs in self.get_data():
            if view.cw_rset is not None:
                if not view.cw_rset:
                    # entity added and deleted in the same transaction
                    # (cache effect)
                    continue
                elif deleted(view.cw_rset[view.cw_row or 0][view.cw_col or 0]):
                    # entity added and deleted in the same transaction
                    continue
            try:
                view.render_and_send(**viewargs)
            except Exception:
                # error in post commit are not propagated
                # We keep this logic here to prevent a small notification error
                # to prevent them all.
                self.exception('Notification failed')


class NotificationHook(hook.Hook):
    __abstract__ = True
    category = 'notification'

    def select_view(self, vid, rset, row=0, col=0):
        try:
            return self._cw.vreg['views'].select_or_none(vid, self._cw, rset=rset,
                                                         row=row, col=col)
        except RegistryNotFound: # can happen in some config
                                 # (e.g. repo only config with no
                                 # notification views registered by
                                 # the instance's cubes)
            return None


class StatusChangeHook(NotificationHook):
    """notify when a workflowable entity has its state modified"""
    __regid__ = 'notifystatuschange'
    __select__ = NotificationHook.__select__ & is_instance('TrInfo')
    events = ('after_add_entity',)

    def __call__(self):
        entity = self.entity
        if not entity.from_state: # not a transition
            return
        rset = entity.related('wf_info_for')
        view = self.select_view('notif_status_change', rset=rset, row=0)
        if view is None:
            return
        comment = entity.printable_value('comment', format='text/plain')
        # XXX don't try to wrap rest until we've a proper transformation (see
        # #103822)
        if comment and entity.comment_format != 'text/rest':
            comment = normalize_text(comment, 80)
        viewargs = {'comment': comment,
                    'previous_state': entity.previous_state.name,
                    'current_state': entity.new_state.name}
        notify_on_commit(self._cw, view, viewargs=viewargs)

class RelationChangeHook(NotificationHook):
    __regid__ = 'notifyrelationchange'
    events = ('before_add_relation', 'after_add_relation',
              'before_delete_relation', 'after_delete_relation')

    def __call__(self):
        """if a notification view is defined for the event, send notification
        email defined by the view
        """
        rset = self._cw.eid_rset(self.eidfrom)
        view = self.select_view('notif_%s_%s' % (self.event,  self.rtype),
                                rset=rset, row=0)
        if view is None:
            return
        notify_on_commit(self._cw, view)


class EntityChangeHook(NotificationHook):
    """if a notification view is defined for the event, send notification
    email defined by the view
    """
    __regid__ = 'notifyentitychange'
    events = ('after_add_entity', 'after_update_entity')

    def __call__(self):
        rset = self.entity.as_rset()
        view = self.select_view('notif_%s' % self.event, rset=rset, row=0)
        if view is None:
            return
        notify_on_commit(self._cw, view)


class EntityUpdatedNotificationOp(hook.SingleLastOperation):
    """scrap all changed entity to prepare a Notification Operation for them"""

    def precommit_event(self):
        # precommit event that creates postcommit operation
        cnx = self.cnx
        for eid in cnx.transaction_data['changes']:
            view = cnx.vreg['views'].select('notif_entity_updated', cnx,
                                            rset=cnx.eid_rset(eid),
                                            row=0)
            notify_on_commit(self.cnx, view,
                    viewargs={'changes': cnx.transaction_data['changes'][eid]})


class EntityUpdateHook(NotificationHook):
    __regid__ = 'notifentityupdated'
    __abstract__ = True # do not register by default
    __select__ = NotificationHook.__select__ & hook.issued_from_user_query()
    events = ('before_update_entity',)
    skip_attrs = set(['modification_date'])

    def __call__(self):
        cnx = self._cw
        if cnx.added_in_transaction(self.entity.eid):
            return # entity is being created
        # then compute changes
        attrs = [k for k in self.entity.cw_edited
                 if not k in self.skip_attrs]
        if not attrs:
            return
        changes = cnx.transaction_data.setdefault('changes', {})
        thisentitychanges = changes.setdefault(self.entity.eid, set())
        rqlsel, rqlrestr = [], ['X eid %(x)s']
        for i, attr in enumerate(attrs):
            var = chr(65+i)
            rqlsel.append(var)
            rqlrestr.append('X %s %s' % (attr, var))
        rql = 'Any %s WHERE %s' % (','.join(rqlsel), ','.join(rqlrestr))
        rset = cnx.execute(rql, {'x': self.entity.eid})
        for i, attr in enumerate(attrs):
            oldvalue = rset[0][i]
            newvalue = self.entity.cw_edited[attr]
            if oldvalue != newvalue:
                thisentitychanges.add((attr, oldvalue, newvalue))
        if thisentitychanges:
            EntityUpdatedNotificationOp(cnx)


# supervising ##################################################################

class SomethingChangedHook(NotificationHook):
    __regid__ = 'supervising'
    __select__ = NotificationHook.__select__ & hook.issued_from_user_query()
    events = ('before_add_relation', 'before_delete_relation',
              'after_add_entity', 'before_update_entity')

    def __call__(self):
        dest = self._cw.vreg.config['supervising-addrs']
        if not dest: # no supervisors, don't do this for nothing...
            return
        if self._call():
            SupervisionMailOp(self._cw)

    def _call(self):
        event = self.event.split('_', 1)[1]
        if event == 'update_entity':
            if self._cw.added_in_transaction(self.entity.eid):
                return False
            if self.entity.e_schema == 'CWUser':
                if not (frozenset(self.entity.cw_edited)
                        - frozenset(('eid', 'modification_date',
                                     'last_login_time'))):
                    # don't record last_login_time update which are done
                    # automatically at login time
                    return False
        self._cw.transaction_data.setdefault('pendingchanges', []).append(
            (event, self))
        return True


class EntityDeleteHook(SomethingChangedHook):
    __regid__ = 'supervisingentitydel'
    events = ('before_delete_entity',)

    def _call(self):
        try:
            title = self.entity.dc_title()
        except Exception:
            # may raise an error during deletion process, for instance due to
            # missing required relation
            title = '#%s' % self.entity.eid
        self._cw.transaction_data.setdefault('pendingchanges', []).append(
            ('delete_entity', (self.entity.eid, self.entity.cw_etype, title)))
        return True



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

import os, os.path as osp, glob
import urllib

from cubicweb.devtools.httptest import CubicWebServerTC


class ETwistHTTPTC(CubicWebServerTC):
    def test_put_content(self):
        data = {'hip': 'hop'}
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        body = urllib.urlencode(data)
        response = self.web_request('?vid=put', method='PUT', body=body)
        self.assertEqual(body, response.body)
        response = self.web_request('?vid=put', method='POST', body=body,
                                    headers=headers)
        self.assertEqual(body, response.body)

if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""tests for notification hooks"""

from cubicweb.devtools.testlib import CubicWebTC


class NotificationHooksTC(CubicWebTC):

    def test_entity_update(self):
        """Check transaction_data['changes'] filled by "notifentityupdated" hook.
        """
        with self.admin_access.repo_cnx() as cnx:
            root = cnx.create_entity('Folder', name=u'a')
            cnx.commit()
            root.cw_set(name=u'b')
            self.assertIn('changes', cnx.transaction_data)
            self.assertEqual(cnx.transaction_data['changes'],
                             {root.eid: set([('name', u'a', u'b')])})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()
